window.kkm = {
	"config":{},
	"cleanup":{},
	"serial":{},
	"bluetooth":{},
	"atol":{},
	"atol54fz":{},
	"shtrikh":{},
	"shtrikh54fz":{},
    "pirit": {},
    "pirit54fz": {},
	"find":{}
};

kkm.init = true;
const ENQ = 0x05;
const ACK = 0x06;
const STX = 0x02;
const ETX = 0x03;
const EOT = 0x04;
const DLE = 0x10;
const NAK = 0x15;

Uint8ClampedArray.prototype.join = function(sep) {
	if (! length) return "empty";
	if (1 == this.length) return this[0];

	if (undefined == sep) sep = ",";

	var res = this[0];
	for (var i=1; i<this.length; i++)
		res += sep + this[i];
	return res;
};

function hx(i) {
	return (i||0).toString(16);
}

function int2bytes (i, arr) {
	const ab = new ArrayBuffer(4);
	const int32view = new Uint32Array(ab);
	const byteview = new Uint8ClampedArray(ab);

	int32view[0] = i;

	var res = arr ? arr : [];
	res[0] = byteview[0];
	res[1] = byteview[1];
	res[2] = byteview[2];
	res[3] = byteview[3];

	return res.reverse();
}

function bytes2int (bytes) {
	var res = bytes[3];
	res += bytes[2] << 8;
	res += bytes[1] << 16;
	res += bytes[0] << 24;

	return res;
}

function short2bytes(s) {
	const a = int2bytes(s, [0,0,0,0]);
	return [a[3], a[2]];
}

function int2BCDbytes (val, arr) {
	var b = val, i=0;
	while(0 < b) {
		var c = b % 100;
		arr[i++] = (c%10) + (16*(c/10|0));
		b = (b/100)|0;
	}
	return arr.reverse();
}

function fromBCD (bytes) {
	var res = 0;
	for (var i=0; i<4; i++) {
		res += (bytes[3-i]%16) * Math.pow(100, i);
		res += (((bytes[3-i]/16)|0)*10) * Math.pow(100, i);
	}

	return res;
}

function encode(cp, str) {
	function code(char) {
		const res = cp.indexOf(char);
		return (-1 == res ? 0x20 : res);
	}

	const bytes = [];
	for (var i=0;i<str.length;i++)
		bytes[i] = code(str[i]);

	return bytes;
}

function cp866() {
	return (
		"␀␁␂␃␄␅␆␇␈␉␊␋␌␍␎␏"+
		"␐␑␒␓␔␕␖␗␘␙␚␛␜␝␞␟"+
		" !\" №%&'()*+,-./"+
		"0123456789:;<=>?"+
		"@ABCDEFGHIJKLMNO"+
		"PQRSTUVWXYZ    -"+
		" abcdefghijklmno"+
		"pqrstuvwxyz{|}~ "+
		"АБВГДЕЖЗИЙКЛМНОП"+
		"РСТУФХЦЧШЩЪЫЬЭЮЯ"+
		"абвгдежзийклмноп"+
		"                "+
		"                "+
		"                "+
		"рстуфхцчшщъыьэюя"+
		"Ёё              "
	);
}

function cp866_pirit() {
	return (
		"␀␁␂␃␄␅␆␇␈␉␊␋␌␍␎␏"+
		"␐␑␒␓␔␕␖␗␘␙␚␛␜␝␞␟"+
		" !\" $%&'()*+,-./"+
		"0123456789:;<=>?"+
		"@ABCDEFGHIJKLMNO"+
		"PQRSTUVWXYZ    -"+
		" abcdefghijklmno"+
		"pqrstuvwxyz{|}~ "+
		"АБВГДЕЖЗИЙКЛМНОП"+
		"РСТУФХЦЧШЩЪЫЬЭЮЯ"+
		"абвгдежзийклмноп"+
		"                "+
		"                "+
		"                "+
		"рстуфхцчшщъыьэюя"+
		"Ёё          №   "
	);
}

function win1251() {
	return (
		"␀␁␂␃␄␅␆␇␈␉␊␋␌␍␎␏"+
		"␐␑␒␓␔␕␖␗␘␙␚␛␜␝␞␟"+
		" !\"  %&'()*+,-./"+
		"0123456789:;<=>?"+
		"@ABCDEFGHIJKLMNO"+
		"PQRSTUVWXYZ    -"+
		" abcdefghijklmno"+
		"pqrstuvwxyz{|}~ "+
		"                "+
		"                "+
		"        Ё       "+
		"        ё№      "+
		"АБВГДЕЖЗИЙКЛМНОП"+
		"РСТУФХЦЧШЩЪЫЬЭЮЯ"+
		"абвгдежзийклмноп"+
		"рстуфхцчшщъыьэюя"
	);
}


function cp866encode (str) {
	return encode(cp866(), str);
}

function cp866encode_pirit (str) {
	return encode(cp866_pirit(), str);
}

function win1251encode(str) {
	return encode(win1251(), str);
}

function decode(cp, bytes) {
	const res = [];
	for (var i=0; i<bytes.length; i++)
		res[i] = cp[bytes[i]];
	return res.join("");
}

function cp866decode (bytes) {
	return decode(cp866(), bytes);
}

function win1251decode (bytes) {
	return decode(win1251(), bytes);
}

function charencode(str) {
    const res = [];
    for (var i=0; i<str.length;i++)
        res[i] = str.charCodeAt(i);
    return res;
}

function chardecode(arr) {
    const res = [];
    for (var i=0; i<arr.length; i++)
        res[i] = String.fromCharCode(arr[i]);
    return res.join("");
}

function fillarr(ch, len) {
	const res = new Array(len);
	for (var i=0;i<len;i++)
		res[i] = ch;
	return res;
}

function zeroarr(len) {
	return fillarr(0, len);
}

function cutbytes(bytes, cut) {
	if (cut <= bytes.length)
		return bytes.slice(0,cut);
	return bytes.concat(zeroarr(cut-bytes.length));
}

function splitbytes(bytes, len) {
	const res = [];
	for (var i=0; i*len+len<bytes.length; i++)
		res[i] = bytes.slice(i*len, i*len+len);
	res.push(bytes.slice(i*len, bytes.length));
	return res;
}

function squinter(left, right, len) {
	const lr = left.length  + right.length;
	return lr > len ? left + right.substr(lr - len) : left + fillarr(" ", len - lr).join("") + right;
}

function center(str, ll, ident) {
	const cut = str.substr(0, ll);
	const i = fillarr(ident, ((ll - cut.length) / 2)|0).join('');
	const res = i+cut+i;
	return res + (ll > res.length ? ident : '');
}

function firstIndex(b, l, p) {
	for (var i=0; i<l; i++)
		if (p(i, b[i])) return i;
	return -1;
}
function lastIndex(b, l, p) {
	var i=l; while(--i>=0)
		if (p(i, b[i]))
			return i;
	return -1;
}
kkm.convert = true;
if (!kkm.util)
	kkm.util = {};

kkm.util.incl = function(module) {
	for (k in module)
		this[k] = module[k];
};

kkm.util.empty = function(f) {
	f();
};

kkm.util.crc = function (data) {
	if (2 > data.length) return 0;
	return data.reduce((a, c) => a ^ c);
};

kkm.util.bytecheck = function (vals) {
	if (!vals || 0 == vals.length)
		return function (b, l) {
			return (0 < l);
		}
	else
		return function (b, l) {
			for (var i=0; i<vals.length; i++)
				if (vals[i] == b[0])
					return true;
			return false;
		}
};

kkm.util.assertmaker = function (errtable, err) {
	const e = (
		errtable ?
		function (c) { return errtable[c]; }:
		function (c) { return c; }
	);

	return function (msg, close, succ) {
		return function (data) {
			const errid = 0xff == data[0] ? 2 : 1;
			if (data[errid])
				close(function () {
					err(msg+" ["+e(data[errid])+"]");
				});
			else if (succ)
				succ(data);
		};
	}
};

kkm.util.map = function (list, proc, res) {
	function iter(num) {
		if (num == list.length)
			return res();
		proc(list[num], function() { iter(num+1); });
	}
	iter(0);
};

kkm.util.ofd = function (enc, line, header, sep) {
	return function(ofdcode, kktnumber, time, offline) {
		if (!ofdcode && !kktnumber) return [];

		var autonom = enc(" (автономный)");
		var tab = enc(fillarr(" ", ((line.length - 10) / 2)));

		return [line, tab.concat(header), line].
			concat(splitbytes(enc("Фиск. чек: "+ofdcode+(offline ? autonom : "")), line.length)).
			concat([enc("Код ККТ ФНС: "+kktnumber)]).
			concat(splitbytes(enc("Время: "+time), line.length)).
			concat(splitbytes(enc("Для проверки Вашего чека зайдите на сайт: http://consumer.ofd-gnivc.ru"), line.length)).
			concat([line]);
	};
};

kkm.util.pempty = function () {
	return Promise.resolve();
}

kkm.util.pass = function(v) {
	return function () {
		return Promise.resolve(v);
	}
}

kkm.util.cbpass = function(data, onfinish) {
	onfinish(data);
}

kkm.util.talkmaker = function(send, receive, log) {
	return function (data) {
		return new Promise(function(res, rej) {
			receive(res);
			send(data, function(){
				log("...sent to fr: ["+data.map(hx)+"]");
			});
		});
	};
}

kkm.util.safetalkmaker = function (send, receive, close, log, assert) {
	return function(data, assertmsg) {
		return new Promise(function (res, rej) {
			receive(assert(assertmsg, close, res));
			send(data, function () {
				log("...safely sent to fr: ["+data.map(hx)+"]");
			});
		});
	};
}

kkm.util.waiter = function (checker, timeout) {
	return new Promise(function (res, rej) {
		function retry() {
			checker().then(function (complete) {
				if (complete) res();
				else setTimeout(retry, timeout);
			});
		}
		retry();
	});
}

kkm.util.modlog = function(logging, prefix)  {
	const l = logging.log;
	return {
		log: function(m) { l(prefix+m); },
		err: logging.err
	};
};

kkm.util.fromstorage = function() {
	return new Promise(function (res) {
		chrome.storage.local.get(res);
	});
};

kkm.util.tostorage = function(items) {
	return new Promise(function (res) {
		chrome.storage.local.set(items, res);
	});
};
kkm.util.tail = function(b, ll, tl) {
	const l = b.length % ll;
	if (l > tl)
		return [splitbytes(b.slice(0, b.length - l), ll), b.slice(l)];

	const el = tl - l;
	const c = el/ll|0;

	const rtl = c*ll + l;

	const s = b.length - rtl;
	return [splitbytes(b.slice(0, s), ll), b.slice(s)];
}

kkm.util.bytes2str = data => "[" + data.map(hx).join(", ") + "]";

kkm.util.chunk = (arr, size) => {
	const copy = arr.slice();
	const chunks = [];

	while (copy.length > 0) {
		chunks.push(copy.slice(0, size));
	}

	return chunks;
};

kkm.util.pjam = function() {
	return new Promise(function(){});
}

kkm.util.fslogging = function(logging, fserr) {
	const L = logging();
	return function () {
		return {
			log: L.log,
			err: fserr(L.log)
		}
	}
}

kkm.util.id = function(v) { return v; }
kkm.config.default = function () {
	function comprefs() {
		return {
			"persistent":false,
			"name":"KKM",
			"bufferSize": 512,
			"bitrate": kkm.config.default.bitrate,
			"dataBits": "eight",
			"parityBit": "no",
			"stopBits": "one",
			"ctsFlowControl": false,
			"sendTimeout": 60000
		};
	}

	function atol() {
		return {
			"vendor": "atol",
			"id": this.id,
			"comprefs": this.comprefs,
			"password": [0, 0],
			"operpass": [0x00, 0x00, 0x00, 0x04],
			"adminpass": [0x00, 0x00, 0x00, 0x30],
			"timeout": this.timeout,
			"cmdtimeout": this.cmdtimeout,
			"errtable": kkm.atol.errors()
		};

	}

	function atol54fz() {
		return {
			"vendor": "atol54fz",
			"id": this.id,
			"comprefs": this.comprefs,
			"password": [0, 0],
			"operpass": [0x00, 0x00, 0x00, 0x04],
			"adminpass": [0x00, 0x00, 0x00, 0x30],
			"timeout": this.timeout,
			"cmdtimeout": this.cmdtimeout,
			"retries":5,
			"errtable": kkm.atol54fz.errors()
		};

	}

	function shtrikh() {
		return {
			"vendor": "shtrikh",
			"id": this.id,
			"comprefs": this.comprefs,
			"password": [0x00, 0x00],
			"operpass": [0x01, 0x00, 0x00, 0x00],
			"adminpass": [0x1e, 0x00, 0x00, 0x00],
			"timeout": this.timeout,
			"cmdtimeout": this.cmdtimeout,
			"porttimeout": this.porttimeout,
			"retries": this.retries,
			"errtable": kkm.shtrikh.errors(),
			"nocutmodels":[
					0,
					6,
					19
				]
		};
	}

	function shtrikh54fz() {
	    const conf = shtrikh.apply(this);
	    conf.vendor = "shtrikh54fz";
	    return conf;
	}

	function atol_bluetooth() {
		return {
			"vendor": "atol_bluetooth",
			"id": this.id,
			"name": this.name,
			"password": [0, 0],
			"operpass": [0x00, 0x00, 0x00, 0x04],
			"adminpass": [0x00, 0x00, 0x00, 0x30],
			"timeout": this.timeout,
			"cmdtimeout": this.cmdtimeout,
			"porttimeout": this.porttimeout,
			"retries": this.retries,
			"errtable": kkm.atol.errors()
		};
	}

    function pirit() {
        return {
            "vendor": "pirit",
            "id": this.id,
            "comprefs": comprefs,
            "password": [0x50, 0x49, 0x52, 0x49],
            "operpass": [0x50, 0x49, 0x52, 0x49],
            "adminpass": [0x50, 0x49, 0x52, 0x49],
            "timeout": this.timeout,
            "cmdtimeout": this.cmdtimeout,
            "porttimeout": this.porttimeout,
            "receiveTimeout": 505,
            "sendTimeout": 500,
            "retries": 5,
            "errtable": kkm.pirit.errors()
        };
    }

    function pirit54fz() {
        const conf = pirit.apply(this);
        conf.vendor = "pirit54fz";
        return conf;
    }


    const res = {
        "id": false,
        "comprefs": comprefs,
        "timeout": 80000,
        "cmdtimeout":15000,
        "porttimeout":500,
        "retries":7,
    };

   	function own(f) {
   		const r = res;
   		return function () {
   			return f.apply(r);
   		}
   	}

    res.atol = own(atol);
    res.atol54fz = own(atol54fz);
    res.shtrikh = own(shtrikh);
    res.shtrikh54fz = own(shtrikh54fz);
    res.pirit = own(pirit);
    res.pirit54fz = own(pirit54fz);
    res.atol_bluetooth = own(atol_bluetooth);

    return res;
};
kkm.config.default.bitrate = 9600;
kkm.config.default.demotext ="Тестовая печать!";

kkm.cleanup.clean = (function () { const selfclean = function () { kkm.cleanup.clean = selfclean; }; return selfclean; })();

kkm.cleanup.add = function (clean) {
	const preclean = kkm.cleanup.clean;
	kkm.cleanup.clean = function() {
		preclean();
		clean();
	};
}
kkm.request = function (logging, net, config, client, oncomplete) {
	net(function (n_connect, n_send, n_flush, n_listener_add, n_err_add, n_close) {
		var connected = false;

		const l = logging();
		const log = function (msg) { l.log("......"+msg) };
		const err = l.err;

		const bin = function(data) {
			var buf = new ArrayBuffer(data.length)
			var view = new Uint8ClampedArray(buf);
			view.set(data);
			return buf;
		}

		const flush = function(conn, onflush) {
			n_flush(conn, function (flushed) {
				log("serial is "+(flushed ? "":"not ")+"flushed"+(flushed ? "" : ", "+(chrome.runtime.lastError || {message:"error unknown"}).message));
				onflush();
			});
		}

		const send1 = function (conn, data, onsend) {
			if (!connected) return;

			log("sending ["+data.map(hx)+"]");

			n_send(conn, bin(data), onsend);
		}

		const receive1 = (function () {
			var conn = false;
			const handlers = [];
			const buff = new Uint8ClampedArray(new ArrayBuffer(512));
			var shift = 0;

			function store(ab) {
				const v = new Uint8ClampedArray(ab);
				for (var i=0; i<v.length; i++)
					buff[shift+i] = v[i];
				shift += v.length;
				log("shift is "+shift);
			}

			function restore() {
				const res = [];
				for (var i=0; i<shift; i++)
					res[i] = buff[i];
				return res;
			}

			function handle() {
				log("searching for appropriate handler for conn "+conn+", data ["+restore().map(hx)+"]");
				const h = handlers[handlers.length-1];
				if (!h) {
					log("no handlers specified for connection "+conn);
					return;
				}

				const matches = h[0](buff, shift);
				log("handler "+(matches ? "": "not ")+"match");
				if (! matches) return;

				handlers.length--;
				const data = restore();
				log("handling "+data.map(n=>(n).toString(16)));
				h[1](data);
				shift = 0;
			}

			const receivehandler = function (c, data) {
				if (! connected) return;

				log("receive");
				if (!conn) {
					log("unexpected receive, terminating it");
					return;
				}
				if (c != conn) {
					log("incorrect connection "+c+" (should be "+conn+")");
					return;
				}

				store(data);
				handle();
			}

			const receiverrhandler = function (msg) {
				log("Ошибка при получении ответа от устройства: "+(msg.error || msg));
			}

			const adder = function (listeners, addlistener, id, caller, errmsg) {
				return function (conn, listener) {
					log("adding listener");
					var virgin = true;
					for (c in listeners) {
						virgin = false;
						break;
					}
					if (virgin)
						addlistener(function (info) {
							if (listeners[id(info)])
								caller(info, listeners[id(info)]);
							else
								log(errmsg+id(info));
						});

					listeners[conn] = listener;

					return listener;
				}
			};


			return function (c, check, onreceive) {
				if (! connected)
					return;
				if (! conn) {
					conn = c;
					n_listener_add(conn, receivehandler, adder);
					n_err_add(conn, receiverrhandler, adder);
					log("registered listeners");
				}
				if (c != conn) {
					log("registering handler for incorrect connection, "+c+" (should be "+conn+")");
					return;
				}
				handlers.push([check, onreceive]);
				log("handlers registered: "+handlers.length);
			}
		})();

		const logconn = function(ci) {
			log("-----------------------");
			log("connection id: "+ci.connectionId);
			log("bufferSize: "+ci.bufferSize);
			log("receiveTimeout: "+ci.receiveTimeout);
			log("sendTimeout: "+ci.sendTimeout);
			log("boadrate: "+ci.bitrate);
			log("-----------------------");
		}

		log("connecting to port "+config.id);
		var concanceled = false;
		const cancel = function(oncomplete) {
			concanceled = function () {
				kkm.request.break.delete(cancel);
				oncomplete();
			};
		};
		kkm.request.break.add(cancel);
		n_connect(config, function (c, logconn) {
			if (concanceled) {
				log("connection halted manualy");
				if (concanceled instanceof Function)
					concanceled();
				return;
			} else
				kkm.request.break.delete(cancel);

			connected = !!c;

			const close = function(next) {
				if (wd) clearTimeout(wd);
				if (!connected) {
					log("connection already closed");
					return;
				}

				flush(c, function () {
					n_close(c, function () {
						kkm.request.break.delete(close);
						connected = false;

						log("-----------------------");
						log("disconnected");
						log("-----------------------");
						if (next && next instanceof Function)
							next();
					});
				});
			}

			kkm.request.break.add(close);

			const state = {stagecritical:true};
			var wd = setTimeout(function () {
					close(function () {
						if (state.stagecritical) err("Устройство перестало отвечать", false);
						else {
							log("timed out on non-critical stage");
							oncomplete();
						}
					});
			}, config.timeout);

			try {
				client(c, send1, receive1, close, flush, state);
			} catch (e) {
				err("Ошибка при обмене данными с устройством: "+e.message);
			}
		}, function () {
			log("connection to "+config.id+" failed, cleanup");
			if (concanceled && concanceled instanceof Function)
				concanceled();
			else
				kkm.request.break.delete(cancel);

			log("kkm.request.break.size "+kkm.request.break.size);
		});
	});
};

kkm.request.break = new Set();
kkm.serial.listener = {};
kkm.serial.err = {};

kkm.serial.request = function (logging, config, client, confail, oncomplete) {
	const U = kkm.util;
	const L = U.modlog(logging(), ".........");

	const log = L.log;
	const err = L.err;

	function logconn(ci) {
		log("-----------------------");
		log("connection id: "+ci.connectionId);
		log("bufferSize: "+ci.bufferSize);
		log("receiveTimeout: "+ci.receiveTimeout);
		log("sendTimeout: "+ci.sendTimeout);
		log("bitrate: "+ci.bitrate);
		log("-----------------------");
	}

	function connect (config, connected, reqfail) {
		function c(cd) {
			chrome.serial.connect(config.id, config.comprefs(), function (ci) {
				const le = chrome.runtime.lastError;
				if (le || (!ci || !ci.connectionId)) {
					log("connection to port "+config.id+" failed with "+le.message);
					if (0 == cd) {
						confail();
						reqfail();
						throw new Error("Ошибка при подключении к порту "+config.id);
					}
					log("retrying");
					setTimeout(function() { c(cd - 1); }, 2000);

				} else {
					logconn(ci);
					connected(ci.connectionId);
				}
			});
		}
		c(3);
	}

	function send(conn, data, sent) {
		chrome.serial.send(conn, data, sent);
	}

	function flush(conn, flushed) {
		chrome.serial.flush(conn, flushed);
	}

	const id = function (info) { return info.connectionId; };

	function listener_add(conn, listener, adder) {
		adder(
			kkm.serial.listener,
			function (l) { chrome.serial.onReceive.addListener(l); },
			id,
			function(info, listener) { listener(info.connectionId, info.data); },
			"no serial listener for connection "
		)(conn, listener);
	}

	function err_add(conn, listener, adder) {
		adder(
			kkm.serial.err,
			function (l) { chrome.serial.onReceiveError.addListener(l) },
			id,
			function (info, listener) { listener(info); },
			"no serial error listener for connection "
		)(conn, listener);
	}

	function close(conn, closed) {
		chrome.serial.disconnect(conn, closed);
	}

	function net(proc) {
		proc(connect, send, flush, listener_add, err_add, close);
	}

	kkm.request(logging, net, config, client, oncomplete);
 }

kkm.serial.enum = function (logging, config, count, proc) {
	const L = kkm.util.modlog(logging(), ".........");
	const log = L.log;
	const err = L.err;

	try {
		chrome.serial.getDevices(function (ports) {
			count(ports.length);
			for (var i=0; i<ports.length; i++)
				try {
					proc(function (config) {
						const conf = config();
						conf.id = ports[i].path;
						conf.name = ports[i].displayName;
						conf.vendor = ports[i].vendorId;
						conf.product = ports[i].product;
						return conf;
					});
				} catch (e) { err(e.message); }
		});
	} catch (e) { err(e.message); }
};
kkm.bluetooth.uuid = "00001101-0000-1000-8000-00805f9b34fb";
kkm.bluetooth.listener = {};
kkm.bluetooth.err = {};

kkm.bluetooth.request = function (logging, config, client, confail) {
	kkm.util.incl(kkm.util.modlog(logging(), "........."));

	function connect (config, connected) {
		chrome.bluetoothSocket.create(function(ci) {
			if (!ci || !ci.socketId) {
				err("Невозможно подключиться к bluetooth-устройству: "+chrome.runtime.lastError.message);
				confail();
			}
			chrome.bluetoothSocket.connect(ci.socketId, config.id, kkm.bluetooth.uuid, function () {
				log("connected to bluetooth device "+config.address+" ("+kkm.bluetooth.uuid+")");
				connected(ci.socketId);
			});
		});
	}

	function send(conn, data, sent) {
		chrome.bluetoothSocket.send(conn, data, sent);
	}

	function flush(conn, flushed) {
		flushed(true);
	}

	const id = function (info) { return info.socketId; };

	function listener_add(conn, listener, adder) {
		adder(
			kkm.bluetooth.listener,
			function (l) { chrome.bluetoothSocket.onReceive.addListener(l); },
			id,
			function(info, listener) { listener(info.socketId, info.data); },
			"no bluetooth listener for connection "
		)(conn, listener);
	}

	function err_add(conn, listener, adder) {
		adder(
			kkm.bluetooth.err,
			function (l) { chrome.bluetoothSocket.onReceiveError.addListener(l); },
			id,
			function (info, listener) { listener(info); },
			"no bluetooth error listener for connection "
		)(conn, listener);
	}

	function close(conn, closed) {
		chrome.bluetoothSocket.disconnect(conn, closed);
		delete kkm.bluetooth.listener[conn];
		delete kkm.bluetooth.err[conn];
	}

	function net(proc) {
		proc(connect, send, flush, listener_add, err_add, close);
	}

	kkm.request(logging, net, config, client);
}

kkm.bluetooth.enum = function (logging, config, count, proc) {
	kkm.util.incl(kkm.util.modlog(logging(), "........."));

	try {
		log("enumerating bluetooth devices");
		if (!chrome.bluetooth) {
			err("bluetooth not supported");
			count(0);
			return;
		}
		chrome.bluetooth.getDevices(function (devs) {
			log("found "+devs.length+" bluetooth devices");
			count(devs.length);
			for (var i=0; i<devs.length; i++) {
				if (-1 != devs[i].uuids.indexOf(kkm.bluetooth.uuid)) {
					log("found known bluetooth kkm");
					try {
						proc(function (config) {
							const conf = config();
							conf.id = devs[i].address;
							log("found kkm address: "+conf.id);
							conf.name = devs[i].name;
							log("found kkm name: "+conf.name);
							return conf;
						});
					} catch (e) { err(e.message); }
				}
			}
		});
	} catch (e) { err(e.message); }
};
if (!kkm.egais)
	kkm.egais = {};

kkm.egais.date = function (d) {
	function f(n) {
		const r = ""+n;
		return 2==r.length?r:"0"+r;
	};

	return d.getFullYear()+"."+f(1+d.getMonth())+"."+f(d.getDate())+" "+f(d.getHours())+":"+f(d.getMinutes());
}

kkm.egais.head = function (e, rec, cashbox, ll) {
	var res = [];
	res.push(e(" "));
	res.push(e(center(rec.head, ll, ' ')));
	res.push(e(" "));
	res.push(e(squinter("ИНН:"+rec.inn, "КПП:"+rec.kpp, ll)));
	res.push(e(squinter("КАССА:"+cashbox, "СМЕНА:"+rec.shift, ll)));
	res.push(e(squinter("ЧЕК:"+rec.doc, "ДАТА:"+kkm.egais.date(new Date()), ll)));
	res.push(e(" "));

	return res;
}

kkm.egais.tail = function (e, rec, ll) {
	var res = [e(" ")];
	res = res.concat(splitbytes(e(rec.url), ll));
	res.push(e(" "));
	res = res.concat(splitbytes(e(rec.sign), ll));
	res.push(e(" "));
	res.push(e(" "));
	res.push(e(" "));
	res.push(e(" "));
	res.push(e(" "));
	res.push(e(" "));

	return res;
}

kkm.egais.parseinfo = function(data) {
	const res = {};
	res.cashbox = fromBCD([0x00, 0x00, data[2], data[3]]);
	res.receipt = fromBCD([0x00, 0x00, data[18], data[19]]);
	res.change = fromBCD([0x00, 0x00, data[20], data[21]]);

	return res;
}
kkm.atol.command = function (config, client, logging, conn, send1, receive1, close, flush) {
	const L = kkm.util.modlog(logging(), '...');
	const log = L.log;
	const err = L.err;
	const bytechecker = kkm.util.bytecheck;

	function esc(data) {
		var i = -1;
		while (++i < data.length)
			if (ETX == data[i] || DLE == data[i])
				data.splice(i++, 0, DLE);
	}

	function unesc (data) {
		var i = 0;
		while(i++<data.length)
			if (data[i] == DLE)
				data.splice(i,1);
		return data;
	}

	function pack(data) {
		esc(data);
		data.push(ETX);
		data.push(kkm.util.crc(data));
		data.unshift(STX);

		return data;
	}

	function shft(b, l) {
		for (var i=0; i<l; i++) {
			if (STX == b[i])
				if (0 == i || DLE != b[i-1])
					return i;
		}
		return -1;
	}

	function unpack(data) {
		const sh = shft(data, data.length);
		if (-1 == sh)
			throw new Error ("no STX found");
		const d = data.slice(sh);
		d.shift();
		if (ETX != d[d.length - 2])
			throw new Error("no ETX found");
		d.pop();
		d.pop();
		return unesc(d);
	}

	function receivedpacket(b, l) {
		log("checking for packet data");
		const sh = shft(b, l);
		log("package shift is "+sh);
		return (0 <= sh) && (3 < (l-sh) && STX == b[sh] && ((ETX == b[l-2] && DLE == b[l-3] && DLE == b[l-4]) || (ETX == b[l-2] && DLE != b[l-3])));
	}

	function r(conn, check, callback) {
		receive1(conn, check, function (d) {
			flush(conn, function (){
				callback(d);
			});
		})
	}

    function sendN(conn, bytes, times, timeout, check, callback) {
        var done = false;
        r(conn, check, function(d) {
            done = true;
            callback(d);
        });
        function attempt(num) {
            if (done) return;
            if (num > times) {
                err("failed sending ["+bytes.map(hx)+"] in "+times+" attempts");
                return;
            }
            log("sending ["+bytes.map(hx)+"], attempt "+num);
            send1(conn, bytes, function() {
                log("attempt "+num+" sent");
                setTimeout(function() { attempt(num + 1); }, timeout);
            });
        }

        attempt(1);
    }

	function send (sdata, sentlink) {
		function checkinit(b) {
			if (ACK == b) {
				log("send init confirmed");
				return true;
			}
			if (ENQ == b || NAK == b)
				send(sdata, sentlink);
			else
				log("incorrect send init, "+b+" insteed of "+ACK);
			return false;
		}

		function checkconfirm(b) {
			if (ACK == b) {
				log("send confirmed");
				return true;
			}
			if (EOT == b || NAK == b)
				send(sdata, sentlink);
			else
				log("incorrect send confirmation, "+b+" insteed of "+ACK);
			return false;
		}

		sendN(conn, [ENQ], 5, 500, bytechecker(), function(d) {
			if (!checkinit(d[0]))
				return;

			log("send pre-confirmed");
			r(conn, bytechecker(), function(d) {
				if (!checkconfirm(d[0]))
					return;

				send1(conn, [EOT], function (si) {
					if (!si || si.error)
						log("failed complete sending data, "+chrome.runtime.lastError.messsage);
					else {
						log("sending completed");
						sentlink(send, receive);
					}
				});
			});
			send1(conn, pack(sdata), function(si) {
				log((!si || si.error) ? "failed sending data, "+chrome.runtime.lastError.message : "data sent!");
			});
		});
	};

	function receive(receivedlink) {
		function check(d) {
			if (ENQ == d[0]) return true;
			if (ACK == d[0]) {
				send1(conn, [ACK], function () {
					receive(receivedlink);
				});
				return false;
			}
			log("incorrect recieve init, "+d[0]+" insteed of "+ENQ);
			return false;
		}

		function checksend(si, fail, ok) {
			if (!si || si.error)
				log(fail+", "+chrome.runtime.lastError.message);
			else
				log(ok);
		}

		r(conn, bytechecker(), function(d) {
			if (!check(d))
				return;

			log("receive initted");
			r(conn, receivedpacket, function(d) {
				const data = unpack(d);
				log("data received! received data size: "+data.length);

				(function finish() {
					r(conn, bytechecker(), function(d) {
						if (EOT != d[0])
							finish();
						else {
							log("receive finished");
							receivedlink(data, send, receive);
						}
					});
				})();
				send1(conn, [ACK], function(si) {
					checksend(si, "failed sending post-receive confirmation", "data receive confirmed");
				});
			});
			send1(conn, [ACK], function (si) {
				checksend(si, "failed sending pre-receive confirmation", "receive pre-confirmation sent");
			});
		});
	}

	client(send, receive, close);
}
kkm.atol.print = {};

kkm.atol.print.linelengths = {
	13:40, 14:20, 15:20, 16:24,
	20:48, 23:39, 24:38, 27:38,
	30:56, 31:32, 32:56, 35:36,
	41:56, 45:56, 46:72, 47:25,
	51:32, 52:32, 53:57
};


kkm.atol.print.printlinemaker = function(stalk) {
	return function(p, l) {
		return p.then(function () {
			return stalk([0x4c].concat(l), "Ошибка при печати строки ("+l+")");
		});
	}
}

kkm.atol.print.modesuper = function (data, log, stalk) {
	const mode = data[17];
	if (1 < mode) {
		log("переходим в режим выбора");
		return stalk([0x48], "Ошибка выхода в режим выбора");
	}
	log("уже в режиме выбора или регистрации");
	return Promise.resolve()
};

kkm.atol.print.modemaker = function(tomode, modename) {
	return function(log, err, talk, stalk, config) {

		return talk([0x3f]).
		then(function(data) {
			const r = data[22];

			if (!r) return Promise.resolve(data);
			return stalk([0x59], "Ошибка аннулирования чека").
			then(function() {
				return talk([0x3f]);
			});
		}).
		then(function(data) {
			const m = data[17];
			if (0 == m || tomode == m) return Promise.resolve(data);

			return stalk([0x48], "Ошибка выхода в режим выбора").
			then(function() {
				return talk([0x3f]);
			});
		}).
		then(function(data) {
			const m = data[17];
			if (tomode == m) return Promise.resolve(data);

			if (0 != m && tomode != m) err("Не удалось выйти в режим выбора");

			return talk([0x56, tomode].concat(config.adminpass), "Ошибка перехода в режим "+tomode).
			then(function() {
				return talk([0x3f]);
			});
		}).
		then(function(data) {
			const m = data[17];
			if (tomode != m) err("Не удалось перейти в режим "+modename);

			return Promise.resolve();
		});
	}
}

kkm.atol.print.moderegister = function(data, log, talk, stalk, config) {
	const mode = data[17];
	if (1 == mode) {
		log("уже в режиме регистрации");
		return Promise.resolve();
	}

	log("переходим в режим регистрации");
	return stalk([0x56, 0x01].concat(config.operpass), "Ошибка при входе в режим регистрации").
	then(function() {
		return talk([0x3f]);
	}).
	then(function (data) {
		if (1 != data[17])
			throw "Устройство не может войти в режим регистрации";
		else
			return Promise.resolve();
	});
}

kkm.atol.print.openchange = function (data, log, stalk, cashier) {
	const change = data[9];
	if (0 == (change & 2)) {
		log("открываем смену");
		return ((!!cashier) ?
				kkm.atol.print.setreqmaker(stalk, log)(1021, cashier) :
				Promise.resolve()).
		then(function () { return stalk([0x9a, 0x00], "Ошибка открытия смены"); });
	} else {
		log("смена уже открыта");
		return Promise.resolve();
	}
}
kkm.atol.print.cancelreceipt = function (data, log, stalk) {
	const receiptnum = data[22];
	if (0 == receiptnum) {
		log("предыдущий чек закрыт");
		return Promise.resolve();
	} else {
		log("аннулируем не закрытый чек");
		return stalk([0x59], "Ошибка аннулирования чека");
	}
}
kkm.atol.print.openreceipt = function (receipt, log, stalk, receipttype) {
	log("открываем чек"+(receipt.return ? " возврата":""));
	return stalk([0x92, 0x00, receipttype(receipt)], "Ошибка открытия чека");
}
kkm.atol.print.receipttype = function (receipt) {
	return receipt.return ? 0x02 : 0x01;
}

kkm.atol.print.corrreceipttype = function (corrreceipt) {
	return corrreceipt.op == 1 ? 0x07 : 0x09;
}

kkm.atol.print.parseinfo = function (data) {
	const res = {};
	res.cashbox = fromBCD([0x00, 0x00, data[2], data[3]]);
	res.receipt = fromBCD([0x00, 0x00, data[18], data[19]]);
	res.change = fromBCD([0x00, 0x00, data[20], data[21]]);

	return res;
}

kkm.atol.print.printlinemaker = function(stalk) {
	return function (p, l) {
		return p.then(function () {
			return stalk([0x4c].concat(l), "Ошибка при печати строки ("+l+")");
		});
	}
}

kkm.atol.print.pos = function (receipt, rl, log, stalk, printline, infodata, linelengths) {
	function sell(a, q) {
		return [0x52, 0x00].concat(a).concat(q).concat([0x01]);
	}

	function ret(a, q) {
		return [0x57, 0x00].concat(a).concat(q);
	}

	log("печать строки "+JSON.stringify(rl));

	const model = infodata[14];
	const amount = int2BCDbytes(rl.price, [0, 0, 0, 0, 0]);
	const quantity = int2BCDbytes(Math.round(rl.quantity*1000), [0, 0, 0, 0, 0]);
	const name = splitbytes(cp866encode(rl.name), (linelengths[model] || 20));

	log("данные подготовлены, amount "+amount.map(n=>(n).toString(16))+", quantity: "+quantity.map(n=>(n).toString(16))+", name: "+name.map(n=>(n).toString(16)));

	var s = (receipt.return ? ret: sell)(amount, quantity);

	return name.reduce(printline, Promise.resolve()).
	then(function () { return stalk(s, "Ошибка при регистрации позиции чека"); });
}

kkm.atol.print.vat = function(rl) {
	return 0x06;
	/*
	function byffd(v1, v2) {
		if ((1 < kkm.atol54fz.ffd) || kkm.atol54fz.ffd105vat) return v2;
		else return v1;
	}

	if (undefined == rl.vat) return byffd(0x04, 0x06);
	switch(rl.vat) {
		case 0: return byffd(0x01, 0x05);
		case 10: return 0x02;
		case 18: return byffd(0x03, 0x01);
		case 110: return byffd(0x05, 0x04);
		case 118: return byffd(0x06, 0x03);
		default: throw ("Неизвестная ставка НДС "+rl.vat);
	}
	*/
}

kkm.atol.print.paytype = function(ptname) {
	switch (ptname){
		case "cash":
			return 0x01; break;
		case "cashless":
			//return ((1 < kkm.atol54fz.ffd) || kkm.atol54fz.ffd105vat) ? 0x02 : 0x04; break;
			return 0x02; break;
		default: 0x01;
	}
}

kkm.atol.print.corrpos = function(sum, log, stalk) {
	if (!sum) {
		log("нет суммы для печати");
		return Promise.resolve();
	}
  log("печать строки чека коррекции (54-ФЗ)");
	const price = int2BCDbytes(sum, [0, 0, 0, 0, 0, 0]);
	const quantity = int2BCDbytes(Math.round(1*1000), [0, 0, 0, 0, 0]);
	const p = [0xE6].
	concat(0x00).
	concat(zeroarr(64)).
	concat(price).
	concat(quantity).
	concat(0x01).
	concat(0x01).
	concat(zeroarr(6)).
	concat(0x04).
	concat(0x00).
	concat(fillarr(0x00, 16)).
	concat(0x00);
	return stalk(p, "Ошибка при печати позиции чека коррекции (ФЗ-54)");
}

kkm.atol.print.pos54fz = function(receipt, rl, log, stalk, printline, infodata, linelengths) {
	function sumdiscount(rl) {
		if (typeof rl.discountAbs != "undefined") {
			return Math.round(rl.discountAbs);
		} else {
			return Math.round(((rl.price * rl.quantity) * rl.discount) / 100);
		}
	}
	const tail = kkm.util.tail;

	log("печать строки (54-ФЗ) "+JSON.stringify(rl));

	const namebytes = cp866encode(rl.name);
	const price = int2BCDbytes(rl.price, [0, 0, 0, 0, 0, 0]);
	const quantity = int2BCDbytes(Math.round(rl.quantity*1000), [0, 0, 0, 0, 0]);
	const d = sumdiscount(rl);
	const discount = int2BCDbytes(d, [0, 0, 0, 0, 0, 0]);
	const discountdirection = d > 0 ? 0x00: 0x01;

	const p = [0xE6].
	concat(0x00).
	concat(cutbytes(namebytes, 64)).
	concat(price).
	concat(quantity).
	concat(0x01).
	concat(discountdirection).
	concat(discount).
	concat(kkm.atol.print.vat(rl)).
	concat(0x00).
	concat(fillarr(0x00, 16)).
	concat(0x00);

	return stalk(p, "Ошибка при печати позиции чека (ФЗ-54)");
}

kkm.atol.print.print4corr = function (config, corrreceipt, onfinish, app, log, err, assert, talkmaker, safetalkmaker) {
	app(
		function (send, receive, close) {
			const U = kkm.util;
			const talk = talkmaker(send, receive, log);
			const stalk = safetalkmaker(send, receive, close, log, assert);

            const assert0 = U.assertmaker(config.errtable, function(msg) { err(msg, 001); });
            const stalk0 = safetalkmaker(send, receive, close, log, assert0);

			const P = kkm.atol.print;
			const setreq = P.setreqmaker(stalk, log);

			return talk([0x3f]).
			then(function(infodata) {
				return P.modesuper(infodata, log, stalk).then(U.pass(infodata));
			}).
			then(function(infodata) {
				return P.moderegister(infodata, log, talk, stalk, config).then(U.pass(infodata));
			}).
			then(function(infodata) {
				return P.openchange(infodata, log, stalk0, corrreceipt.cashier).then(U.pass(infodata));
			}).
			then(function(infodata) {
				return P.cancelreceipt(infodata, log, stalk).then(U.pass(infodata));
			}).
			then(function(infodata) {
				log("Открываем чек коррекции");
				return stalk([0x92, 0x00, P.corrreceipttype(corrreceipt)], "Ошибка открытия чека коррекции");
			}).
			then(function(infodata) {
				return (!!corrreceipt.cashier) ?
					setreq(1021, corrreceipt.cashier).then(U.pass(infodata)) :
					Promise.resolve(infodata);
			}).
			then(function(infodata) {
				return P.corrpos(corrreceipt.cashsum+corrreceipt.nocashsum,log,stalk);
			}).
			then(function() {
				return P.printsum(corrreceipt.cashsum, P.paytype("cash"), log, stalk);
			}).
			then(function() {
				return P.printsum(corrreceipt.nocashsum, P.paytype("cashless"), log, stalk);
			}).
			then(function(){log("Закрываем чек коррекции");return stalk([0x4a, 0x00].concat(P.corrreceipttype(corrreceipt)).concat([0, 0, 0, 0, 0]), "Ошибка закрытия чека");}).
			then(function(){log("Проверяем состояния ус-ва");return talk([0x3f]);}).
			then(function(data) {
				if (1 == data[22]) {
					return stalk([0x59], "Ошибка аннулирования чека").then(function(){return new Promise(function(res) { close(res); });});
				}
				else {
					return new Promise(function(res) { close(res); });
				}
			}).
			then(onfinish).
			catch(function(e) { close(function() { err(e.message || e); }); });
		}
	);
}

kkm.atol.print.print4pos = function(config, receipt, pos, onfinish, app, log, err, assert, talkmaker, safetalkmaker) {
	app(
		function (send, receive, close) {
			const U = kkm.util;
			const talk = talkmaker(send, receive, log);
			const stalk = safetalkmaker(send, receive, close, log, assert);

			const assert0 = U.assertmaker(config.errtable, function(msg) { err(msg, 001); });
			const stalk0 = safetalkmaker(send, receive, close, log, assert0);

			const P = kkm.atol.print;
			const printline = P.printlinemaker(stalk);
			const egaisprint = P.egaisprintmaker(log, stalk, kkm.egais, P.linelengths, printline);
			const setreq = P.setreqmaker(stalk, log);

			return talk([0x3f]).
			then(function(infodata) {
				return P.modesuper(infodata, log, stalk).then(U.pass(infodata));
			}).
			then(function(infodata) {
				return P.moderegister(infodata, log, talk, stalk, config).then(U.pass(infodata));
			}).
			then(function(infodata) {
				return P.openchange(infodata, log, stalk0, receipt.cashier).then(U.pass(infodata));
			}).
			then(function(infodata) {
				return P.cancelreceipt(infodata, log, stalk).then(U.pass(infodata));
			}).
			then(function(infodata) {
				return P.openreceipt(receipt, log, stalk, P.receipttype).then(U.pass(infodata));
			}).
			then(function(infodata) {
				return (!!receipt.cashier) ?
					setreq(1021, receipt.cashier).then(U.pass(infodata)) :
					Promise.resolve(infodata);
			}).
			then(function(infodata) {
				return (!!receipt.counterparty) ?
					setreq(1008, receipt.counterparty).then(U.pass(infodata)) :
					Promise.resolve(infodata);
			}).
			then(function(infodata) {
				const model = infodata[17];
				const ofdblock = U.ofd(
					cp866encode,
					fillarr(0x2d, (P.linelengths[model] || 20)),
					[0x2a, 0x2a, 0x2a, 0x20, 0x94, 0x8f, 0x20, 0x2a, 0x2a, 0x2a],
					[0x24, 0x20]
				);

				const block = ofdblock(receipt.ofdCode, receipt.ofdKKT, receipt.ofdTime, false);
				return block.reduce(printline, Promise.resolve()).then(U.pass(infodata));
			}).
			then(function(infodata) {
				function sellchain(prev, line) {
					return prev.then(function() { return pos(receipt, line, log, stalk, printline, infodata, P.linelengths); });
				}

				return receipt.positions.reduce(sellchain, Promise.resolve());
			}).
			then(function() {
				return P.regdiscount(receipt, log, stalk);
			}).
			then(function() {
				return P.printsum(receipt.paidSum, P.paytype("cash"), log, stalk);
			}).
			then(function() {
				return P.printsum(receipt.noCashSum, P.paytype("cashless"), log, stalk);
			}).
			then(P.closereceiptmaker(close, talk, stalk, log, receipt, egaisprint, P.pempty)).
			then(onfinish).
			catch(function(e) { close(function() { err(e.message || e); }); });
		}
	);
}

kkm.atol.print.egaisprintmaker = function (log, stalk, E, linelengths, printline) {
	const P = kkm.atol.print;

	return function (er, infodata) {
		function qr() {
			if (!kkm.atol54fz.qr_supported) {
				log("fr is not able to print qr code");
				return Promise.resolve();
			}

			log("printing qr code, text "+er.url);
			return stalk(
						[0xC1,
						0,
						2,
						8,
						0, 0,
						0, 1,
						0,
						0,
						0,
						0, 0,
						0, 0].
						concat(cutbytes(win1251encode(er.url), 100)),
						"Ошибка печати QR-кода"
				);
		}

		return function() {
			const model = infodata[14];
			const head = E.head(cp866encode, er, E.parseinfo(infodata).cashbox, linelengths[model] || 20);
			const tail = E.tail(cp866encode, er, linelengths[model] || 20);

			log("printing egais receipt head, "+head.length+" lines");

			return head.reduce(printline, Promise.resolve()).
			then(qr).
			then(function () {
				return tail.reduce(printline, Promise.resolve());
			});
		}
	}
}

kkm.atol.print.printsum = function(sum, paytype, log, stalk) {
	if (!sum) {
		log("суммы по типу оплаты "+paytype+" нет");
		return Promise.resolve();
	}
	log("печать суммы по типу оплаты "+paytype);
	const sumbytes = int2BCDbytes(sum|0, [0, 0, 0, 0, 0]);
	return stalk([0x99, 0x00].concat(paytype).concat(sumbytes), "Ошибка при печати суммы");
}

kkm.atol.print.regdiscount = function(receipt, log, stalk) {
	if (!receipt.discount) {
		log("скидки нет");
		return Promise.resolve();
	}
	log("регистрируем скидку")
	const discount = int2BCDbytes((receipt.discount)|0, [0, 0, 0, 0, 0]);
	return stalk([0x43, 0x00, 0x00, 0x01, 0x00].concat(discount), "Ошибка регистрации скидки");
}

kkm.atol.print.closereceiptmaker = function (close, talk, stalk, log, receipt, egaisprint, pempty) {
	return function () {
		function closepromise() {
			return new Promise(function(res, rej) { close(res); });
		}

		function paytype() {
			//return ("cash" == receipt.type ? [0x01] : (((1 < kkm.atol54fz.ffd) || kkm.atol54fz.ffd105vat) ? [0x02] : [0x04]));
			return ("cash" == receipt.type ? [0x01] : [0x02]);
		}

		log("закрываем чек");

		const U = kkm.util;
		return stalk([0x4a, 0x00].concat(paytype()).concat([0, 0, 0, 0, 0]), "Ошибка закрытия чека").
		then(function() { return talk([0x3f]); }).
		then(function(data) {
			if (1 == data[22]) {
				return stalk([0x59], "Ошибка аннулирования чека").then(closepromise);
			} else {
				const egais1 = receipt.egais1 ? egaisprint(receipt.egais1, data) : U.pempty;
				const egais2 = receipt.egais2 ? egaisprint(receipt.egais2, data) : U.pempty;

				return egais1().then(egais2).then(closepromise);
			}
		})
	}
}

kkm.atol.print.setreqmaker = function(stalk, log) {
	function bytes(i) {
		const a = int2bytes(i, [0,0,0,0]);
		return [a[3], a[2]];
	}

	return function(req, text) {
		log("setting requisite "+req+" ("+text+")");
		const textbytes = cp866encode(text);
		const reqbytes = bytes(req);
		const lenbytes = bytes(textbytes.length);
		return stalk([
			0xe8,
			0x01,
			0x01,
			0x00
			].concat(reqbytes).
			concat(lenbytes).
			concat(textbytes),
			"Ошибка при установке значения реквизита "+req);
	}
}

kkm.atol.print.setfieldmaker = function (log, err, talk, stalk, config) {
	const AP = kkm.atol.print;
	const modeprog = AP.modemaker(4, "программирования");

	return function(table, row, field, bytes) {
		const rb = int2BCDbytes(row, [0,0,0,0]);
		const progbytes = [0x50, table, rb[2], rb[3], field].concat(bytes);

		return modeprog(log, err, talk, stalk, config).
		then(function() {
			return talk(progbytes);
		});
	}
}
kkm.atol.api = function (logging, config, executor, proc) {
	const U = kkm.util;
	const L = logging();
	const log = L.log;
	const err = L.err;
	const assert = U.assertmaker(config.errtable, err);
	const assert0 = U.assertmaker(config.errtable, function(msg) { err(msg, 001); });

	function p(cmd) {
		return ((config.password || [0,0]).concat(cmd));
	}

	function talkmaker(send, receive, log) {
		const talk = U.talkmaker(send, receive, log);
		return function(data) {
			return talk(p(data));
		}
	};

	function safetalkmaker(send, receive, close, log, assert) {
		const stalk = U.safetalkmaker(send, receive, close, log, assert);
		return function(data, assertmsg) {
			return stalk(p(data), assertmsg);
		}
	};


	function app(chain) {
		executor(logging, chain);
	}

	function beep() {
		app(
			function (send, receive, close) {
				send(p([0x47]), function() {
					log("beeeeep!");
					close();
				});
			}
		);
	}

	function demotext(data, onfinish) {
		log("demo print");
		app(
			function (send, receive, close) {
				log("sending demoprint command");
				function printline(p, l) {
					return p.then(function () {
						return stalk([0x4c].concat(l), "Ошибка при печати строки ("+l+")");
					});
				}
				const stalk = safetalkmaker(send, receive, close, log, assert);
				log("sending demoprint command");
				const mas = [kkm.config.default.demotext,"","","","",""];
				const text = mas.map(s => cutbytes(cp866encode(s),20));
				stalk([0x048],"ошибка при переходе в режим выбора").then
				(
					function()
					{
						return text.reduce(printline, Promise.resolve());
					}
				)
				.then
				(
					function ()
					{
						close();
						if (onfinish && onfinish instanceof Function)
							onfinish();
					}
				);
			}
		);
	}

	function demo(data, onfinish) {
		log("demo print");
		app(
			function (send, receive, close) {
				log("sending demoprint command");
				receive(assert("ошибка при переходе в режим выбора", close, function (data) {
					log("перешли в режим выбора");
					receive(assert("Ошибка при тестовом прогоне", close, function (data) {
						log("got response: "+data.join());
						close();
						if (onfinish && onfinish instanceof Function)
							onfinish();
					}));
					send(p([0x82, 0x01, 0x00, 0x00]), function(send, receive) {
						log("demo print processing!");
					});
				}));
				send(p([0x048]), function () {
					log("переходим в режим выбора");
				});
			}
		);
	}

	function info(proc) {
		app(
			function (send, receive, close) {
				send(p([0x3f]), function() {
					log("info request sent!");
					receive(function (data) {
						close(function () { proc(data); });
					});
				});
			}
		);
	}

	function type(proc) {
		app(
			function (send, receive, close) {
				send(p([0xa5]), function() {
					log("type request sent!");
					receive(function (data) {
						log(data[0] ? "error "+data[0] : "type data received");
	                    close(function () { proc(data); });
					});
				});
			}
		);
	}

	function check(type, info) {
		app(
			function (send, receive, close) {
				send(p([0xa5]), function() {
					log("type request sent!");
					receive(function (data) {
						log(data[0] ? "error "+data[0] : "type data received");
	                    type(data);
						send(p([0x3f]), function() {
							log("info request sent!");
							receive(function (data) {
								log("info data received! ["+data+"]");
								close(function () { info(data); });
							});
						});
					});
				});
			}
		);
	}

	function print(receipt, onfinish) {
		var model;

		function prepare(send, receive, close, data) {
			const talk = talkmaker(send, receive, log);
			const stalk = safetalkmaker(send, receive, close, log, assert);
			const stalk0 = safetalkmaker(send, receive, close, log, assert0);

			model = data[14];
			const mode = data[17];
			const change = data[9];
			const receiptnum = data[22];

			function modesuper() {
				if (1 < mode) {
					log("переходим в режим выбора");
					return stalk([0x48], "Ошибка выхода в режим выбора");
				}
				log("уже в режиме выбора или регистрации");
				return Promise.resolve()
			}

			function moderegister() {
				if (1 == mode) {
					log("уже в режиме регистрации");
					return Promise.resolve();
				}

				log("переходим в режим регистрации");
				return stalk([0x56, 0x01].concat(config.operpass), "Ошибка при входе в режим регистрации").
				then(function() {
					return talk([0x3f]);
				}).
				then(function (data) {
					if (1 != data[17])
						throw "Устройство не может войти в режим регистрации";
					else
						return Promise.resolve();
				});
			}

			function openchange() {
				if (0 == (change & 2)) {
					log("открываем смену");
					return stalk0([0x9a, 0x00], "Ошибка открытия смены");
				} else {
					log("смена уже открыта");
					return Promise.resolve();
				}
			}

			function openreceipt() {
				log("открываем чек"+(receipt.return ? " возврата":""));
				return stalk([0x92, 0x00, receipttype()], "Ошибка открытия чека");
			}

			function cancelreceipt(f) {
				if (0 == receiptnum) {
					log("предыдущий чек закрыт");
					Promise.resolve();
				} else {
					log("аннулируем не закрытый чек");
					return stalk([0x59], "Ошибка аннулирования чека");
				}
			}

			function receipttype() {
				return receipt.return ? 0x02 : 0x01;
			}

			return modesuper().
			then(moderegister).
			then(openchange).
			then(cancelreceipt).
			then(openreceipt);
		}


		const linelengths = {
			13:40, 14:20, 15:20, 16:24,
			20:48, 23:39, 24:38, 27:38,
			30:56, 31:32, 32:56, 35:36,
			41:56, 45:56, 46:72, 47:25,
			51:32, 52:48, 53:57
		};

        const cutEgaisCheckModels = [52];

		const ofdblock = U.ofd(
			cp866encode,
			fillarr(0x2d, (linelengths[model] || 20)),
			[0x2a, 0x2a, 0x2a, 0x20, 0x94, 0x8f, 0x20, 0x2a, 0x2a, 0x2a],
			[0x24, 0x20]
		);

		function paytype() {
			return ("cash" == receipt.type ? [0x01] : [0x04]);
		}

		function parseinfo(data) {
			const res = {};
			res.cashbox = fromBCD([0x00, 0x00, data[2], data[3]]);
			res.receipt = fromBCD([0x00, 0x00, data[18], data[19]]);
			res.change = fromBCD([0x00, 0x00, data[20], data[21]]);

			return res;
		}

        function canCutEgaisCheck(infodata) {
            const model = infodata[14];
            return cutEgaisCheckModels.includes(model);
        }

		app(
			function (send, receive, close) {
				const talk = talkmaker(send, receive, log);
				const stalk = safetalkmaker(send, receive, close, log, assert);

				function printline(p, l) {
					return p.then(function () {
						return stalk([0x4c].concat(l), "Ошибка при печати строки ("+l+")");
					});
				}

				function pos(rl) {
					function sell(a, q) {
						return [0x52, 0x00].concat(a).concat(q).concat([0x01]);
					}

					function ret(a, q) {
						return [0x57, 0x00].concat(a).concat(q);
					}

					log("печать строки "+JSON.stringify(rl));

					const amount = int2BCDbytes(rl.price, [0, 0, 0, 0, 0]);
					const quantity = int2BCDbytes(Math.round(rl.quantity*1000), [0, 0, 0, 0, 0]);
					const name = splitbytes(cp866encode(rl.name), (linelengths[model] || 20));

					log("данные подготовлены, amount "+amount.join()+", quantity: "+quantity.join()+", name: "+name.join());

					var s = (receipt.return ? ret: sell)(amount, quantity);

					return name.reduce(printline, Promise.resolve()).
					then(function () { return stalk(s, "Ошибка при регистрации позиции чека"); });
				}

				function egaisprint(er, infodata) {
					function qr() {
                        if (kkm.atol.qr_supported) {
                            log("printing qr code, text "+er.url);
                            return stalk(
                                [0xC1,
                                    0,
                                    2,
                                    8,
                                    0, 0,
                                    0, 1,
                                    0,
                                    0,
                                    0,
                                    0, 0,
                                    0, 0].
                                    concat(cutbytes(win1251encode(er.url), 100)),
                                "Ошибка печати QR-кода"
                            );
                        } else {
                            log("fr is not able to print qr code");
                            return Promise.resolve();
                        }
					}

                    function cutEgaisCheck () {
                        if (!canCutEgaisCheck(infodata)) {
                            log("fr is not able to cut egais check");
                            return Promise.resolve();
                        }

                        log("cut egais check");
                        return talk([0x75, 0x00]);
                    }

					return function() {
						const E = kkm.egais;
						const head = E.head(cp866encode, er, parseinfo(infodata).cashbox, linelengths[model] || 20);
						const tail = E.tail(cp866encode, er, linelengths[model] || 20);

						log("printing egais receipt head, "+head.length+" lines");

						return head.reduce(printline, Promise.resolve()).
						then(qr).
						then(function () {
							return tail.reduce(printline, Promise.resolve());
						}).
                        then(function () {
                            return stalk([0x6C]);
                        }).
                        then(cutEgaisCheck);
					}
				}

				const block = ofdblock(receipt.ofdCode, receipt.ofdKKT, receipt.ofdTime, false);

				function closereceipt() {
					function closepromise() {
						return new Promise(function(res, rej) { close(res); });
					}

					log("закрываем чек");

					return stalk([0x4a, 0x00].concat(paytype()).concat([0, 0, 0, 0, 0]), "Ошибка закрытия чека").
					then(function() { return talk([0x3f]); }).
					then(function(data) {
						if (1 == data[22]) {
							return stalk([0x59], "Ошибка аннулирования чека").then(closepromise);
						} else {
							const egais1 = receipt.egais1 ? egaisprint(receipt.egais1, data) : U.pempty;
							const egais2 = receipt.egais2 ? egaisprint(receipt.egais2, data) : U.pempty;

							return egais1().then(egais2).then(closepromise);
						}
					})
				}

				function printsum(sum, paytype) {
					if (!sum) {
						log("суммы по типу оплаты "+paytype+" нет");
						return Promise.resolve();
					}
					log("печать суммы по типу оплаты "+paytype);
					const sumbytes = int2BCDbytes(sum|0, [0, 0, 0, 0, 0]);
					return stalk([0x99, 0x00].concat(paytype).concat(sumbytes), "Ошибка при печати суммы");
				}

				function regdiscount() {
					if (!receipt.discount) {
						log("скидки нет");
						return Promise.resolve();
					}
					log("регистрируем скидку")
					const discount = int2BCDbytes((receipt.discount)|0, [0, 0, 0, 0, 0]);
					return stalk([0x43, 0x00, 0x00, 0x01, 0x00].concat(discount), "Ошибка регистрации скидки");
				}

				return talk([[0x3f]]).
				then(function (data) { return prepare(send, receive, close, data); }).
				then(function() { return block.reduce(printline, Promise.resolve()); }).
				then(function() {
					function sellchain(prev, line) {
						return prev.then(function() { return pos(line); });
					}

					return receipt.positions.reduce(sellchain, Promise.resolve());
				}).
				then(function() { return regdiscount(); }).
				then(function() { return printsum(receipt.paidSum, 1); }).
				then(function() { return printsum(receipt.noCashSum, 4); }).
				then(closereceipt).
				then(onfinish);
			}
		);
	}

	function z(data, onfinish) {
		function reportmode(send, receive, close, succ) {
			function modeswitch() {
				log("переход в режим отчетов с гашением")
				receive(assert("Ошибка перехода в режим отчетов с гашением", close, succ));
				send(p([0x56, 0x03]).concat(config.adminpass), function () {});
			}

			function modeexit() {
				log("Ошибка при переходе в режим выбора");
				receive(assert("Ошибка выхода в режим выбора", close, function () {
					modeswitch(succ);
				}));
				send(p([0x48]), function(){});
			}

			receive(function (data) {
				var proc = modeexit;
				if (0 == data[17])
					proc = modeswitch;
				if (3 == data[17])
					proc = U.empty;

				if (0 == (data[9] & 2)) {
					err("Невозможно повторное снятие Z-отчета");
					close(function () {
						if (onfinish)
							onfinish();
					});
				}
				else
					proc(succ);
			});
			send(p([0x3f]), function () {});
		}

		function wait4finish(send, receive, onfinish) {
			log("проверяем завершена ли печать отчета");
			receive(function (data) {
				log("состояние полученo");
				log("в режиме "+data[1]);
				if (0x17 == data[1] || [0x23] == data[1])
					setTimeout(function () { wait4finish(send, receive, onfinish); }, 2000);
				else
					onfinish();
			});
			send(p([0x45]), function () {
				log("отправлен запрос состояния устройства");
			});
		}

		app(
			function(send, receive, close) {
				reportmode(send, receive, close, function() {
					receive(assert("Ошибка запроса Z-отчета", close, function () {
						wait4finish(send, receive, function () {
							close(function () {
								if (onfinish)
									onfinish();
							});
						});
					}));
					send(p([0x5a]), function () { log("Z-отчет запрошен"); });
				})
			}
		);
	}


	function x(data, onfinish) {
		function reportmode(send, receive, close, succ) {
			function modeswitch(succ) {
				log("переход в режим отчетов без гашения")
				receive(assert("Ошибка перехода в режим отчетов без гашения", close, succ));
				send(p([0x56, 0x02]).concat(config.adminpass), function () {});
			}

			function modeexit(succ) {
				log("Ошибка при переходе в режим выбора");
				receive(assert("Ошибка выхода в режим выбора", close, function () {
					modeswitch(succ);
				}));
				send(p([0x48]), function(){
				});
			}

			receive(function (data) {
				var proc = modeexit;
				if (0 == data[17])
					proc = modeswitch;
				if (3 == data[17])
					proc = U.empty;

				proc(succ);
			});
			send(p([0x3f]), function () {
			});
		}

		function wait4finish(send, receive, onfinish) {
			log("проверяем завершена ли печать отчета")
			send(p([0x45]), function () {
				log("отправлен запрос состояния устройства");
				receive(function (data) {
					log("состояние полученo");
					log("в режиме "+data[1]);
					if (0x17 == data[1] || [0x22] == data[1]) {
						log("x-отчет все еще печатается")
						setTimeout(function () { wait4finish(send, receive, onfinish); }, 2000);
					} else {
						onfinish();
					}
				});
			});
		}

		app(
			function(send, receive, close) {
				reportmode(send, receive, close, function() {
					receive(assert("Ошибка запроса X-отчета", close, function () {
						log("х-отчет начал печататься");
						wait4finish(send, receive, function () {
							close(function () {
								if (onfinish)
									onfinish();
							});
						});
					}));
					send(p([0x67, 0x01]), function () { log("X-отчет запрошен"); });
				})
			}
		);
	}

	function moderegister(stalk, talk, checkchange) {
		return talk([0x3f]).
		then(function (data) {
			function reg() {
				return stalk([0x56, 0x01].concat(config.adminpass), "Ошибка при переходе в режим регистрации");
			}

			if (!(checkchange ^ (0 != (data[9] & 2)))) {
				return Promise.reject(new Error(checkchange ? "Смена уже открыта" : "Смена закрыта. Операция невозможна."));
			}

			if (1 == data[17])
				return Promise.resolve();

			if (0 == data[17])
				return reg();

			return stalk([0x48], "Ошибка при переходе в режим выбора").then(reg);
		});
	}

	function change(data, onfinish) {
		app(function(send, receive, close) {
			const stalk0 = safetalkmaker(send, receive, close, log, assert0);
			const talk = talkmaker(send, receive, log);

			moderegister(stalk0, talk, true).
			then(function() {
				return stalk0([0x9a, 0x00], "Ошибка при открытии смены");
			}).
			then(function() { close(onfinish);}, function() { close(onfinish);}).
			catch(function(e) { close(function () { err(e.message || e, 001); }); });
		});
	}

	function cashin(data, onfinish) {
		const sum = data.sum || data;
		app(function(send, receive, close) {
			const stalk = safetalkmaker(send, receive, close, log, assert);
			const talk = talkmaker(send, receive, log);

			moderegister(stalk, talk, false).
			then(function () {
				const sumbytes = int2BCDbytes(sum|0, [0,0,0,0,0]);
				return stalk([0x49, 0x00].concat(sumbytes), "Ошибка при внесении наличных");
			}).
			then(function() { close(onfinish); }).
			catch(function(e) { close(err(e.message ? e.message : e)); });
		});
	}

	function cashout(data, onfinish) {
		const sum = data.sum || data;
		app(function(send, receive, close) {
			const stalk = safetalkmaker(send, receive, close, log, assert);
			const talk = talkmaker(send, receive, log);

			moderegister(stalk, talk, false).
			then(function () {
				const sumbytes = int2BCDbytes(sum|0, [0,0,0,0,0]);
				return stalk([0x4f, 0x00].concat(sumbytes), "Ошибка при выплате наличных");
			}).
			then(function() { close(onfinish); }).
			catch(function(e) { close(err(e.message ? e.message : e)); });
		});
	}

	proc(beep, demotext, info, type, check, print, z, x, change, cashin, cashout);
}
kkm.atol.ext = function (logging, config, executor, proc) {
	kkm.atol.api(logging, config, executor, function (beep, demo, info, type, check, print, z, x, change, cashin, cashout) {
		const U = kkm.util;
		const L = logging();
		const log = L.log;
		const err = L.err;
		const assert = U.assertmaker(config.errtable, err);

		function p(cmd) {
			return ((config.password || [0,0]).concat(cmd));
		}

		function talkmaker(send, receive, log) {
			const talk = U.talkmaker(send, receive, log);
			return function(data) {
				return talk(p(data));
			}
		};

		function safetalkmaker(send, receive, close, log, assert) {
			const stalk = U.safetalkmaker(send, receive, close, log, assert);
			return function(data, assertmsg ){
				return stalk(p(data), assertmsg);
			}
		};


		function app(chain) {
			executor(logging, chain);
		}

		function checkext(res, checked) {
			const datas = [];

			function fsapp(chain) {
				const fslogging = U.fslogging(logging, function(l) { return function (e) { if (e) l(e.message || e); checked(res, datas); }; });
				return executor(fslogging, chain);
			}

			res.vendor = "atol";
			app(function (send, receive, close) {
					receive(function (data) {
						log(data[0] ? "error "+data[0] : "type data received");
	                    datas.push(data);
						res.model = cp866decode(data.slice(11, data.length));
						receive(function (data) {
						    log("info data received! ["+data+"]");
							datas.push(data);
							if(data[15] && data[16]) res.ver = " v. "+cp866decode([data[15],data[16]]).split("").join(".");
							else res.ver = "";
							res.fiscal = (data[9] & 1);
					        res.doc = 0;
					        res.receipt = fromBCD([0x00, 0x00, data[18], data[19]]);
					        res.change = fromBCD([0x00, 0x00, data[20], data[21]]);
					        res.serial = fromBCD([data[13], data[12], data[11], data[10]]);
							res.qrsupported = false;
                			close(function () {
                				fsapp(function (fsend, fsreceive, fsclose) {
                					var read = false;
                					setTimeout(function() {
                						if (read) return;
                						log("ffd version read timed out");
                						fsclose(function() { checked(res, datas); });
                					}, 500);
									fsreceive(function(data) {
										log("ffd version received");
										res.ffd = data[2]; /*-*/ res.ffd = 1.05;
										log("ffd version is "+res.ffd);
										fsreceive(function(data) {
											const ffd105vat = data[2] & 8;
											log("vat table ffd version is "+(ffd105vat ? "1.05" : "1.0"));
											res.ffd105vat = ffd105vat; /*-*/ res.ffd105vat = 1;
											fsclose(function() { checked(res, datas); });
										});
										fsend([0x91, 0x38, 0x00, 0x00], function () {
											log("fr flags request sent!");
										});
	                				});
									fsend(p([0x91, 0x36, 0x00, 0x00]), function() {
										log("ffd request sent!");
									});
                				});
							});
						});
						send(p([0x3f]), function() {
							log("info request sent!");
						});
					});
					send(p([0xa5]), function() {
						log("type request sent!");
					});
				}
			);
		}

		proc(beep, demo, info, type, checkext, print, z, x, change, cashin, cashout, U.cbpass, U.cbpass, U.cbpass);
	});
}
kkm.atol.errors = function () {
	return [
		"Ошибок нет",
		"Контрольная лента обработана без ошибок",
		"",
		"",
		"",
		"",
		"",
		"",
		"Неверная цена (сумма)",
		"",
		"Неверное количеств",
		"Переполнение счетчика наличности",
		"Невозможно сторно последней операции",
		"Сторно по коду невозможно (в чеке зарегистрировано меньшее количество товаров с указанным кодом)",
		"Невозможен повтор последней операции",
		"Повторная скидка на операцию невозможна",
		"Скидка/надбавка на предыдущую операцию невозможна",
		"Неверный код товара",
		"Неверный штрихкод товара",
		"Неверный формат",
		"Неверная длина",
		"",
		"ККТ заблокирована в режиме ввода даты",
		"Требуется подтверждение ввода даты",
		"Нет больше данных для передачи ПО ККТ",
		"Нет подтверждения или отмены продажи",
		"Отчет с гашением прерван. Вход в режим невозможен.",
		"Отключение контроля наличности невозможно (не настроены необходимые типы оплаты).",
		"",
		"",
		"Вход в режим заблокирован",
		"Проверьте дату и время",
		"Дата и время в ККТ меньше чем в ЭКЛЗ/ФП",
		"Невозможно закрыть архив",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"Товар не найден",
		"Весовой штрихкод с количеством <>1.000",
		"Переполнение буфера чека",
		"Недостаточное количество товара",
		"Сторнируемое количество больше проданного",
		"Заблокированный товар не найден в буфере чека",
		"Данный товар не продавался в чеке, сторно невозможно",
		"Memo Plus TM 3 TM заблокировано с ПК",
		"Ошибка контрольной суммы таблицы настроек Memo Plus TM 3 TM",
		"Неверная команда от ККТ",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"Команда не реализуется в данном режиме ККТ",
		"Нет бумаги",
		"Нет связи с принтером чеков",
		"Механическая ошибка печатающего устройства",
		"Неверный тип чека",
		"Нет больше строк картинки",
		"Неверный номер регистра",
		"Недопустимое целевое устройство",
		"Нет места в массиве картинок",
		"Неверный номер картинки / картинка отсутствует",
		"Сумма сторно больше, чем было получено данным типом оплаты",
		"Сумма не наличных платежей превышает сумму чека",
		"Cумма платежей меньше суммы чека",
		"Накопление меньше суммы возврата или аннулирования",
		"Переполнение суммы платежей",
		"",
		"",
		"",
		"",
		"",
		"Данная модель ККТ не может выполнить команду",
		"Неверная величина скидки / надбавки",
		"Операция после скидки / надбавки невозможна",
		"Неверная секция",
		"Неверный вид оплаты",
		"Переполнение при умножении",
		"Операция запрещена в таблице настроек",
		"Переполнение итога чека",
		"Открыт чек аннулирования – операция невозможна",
		"",
		"Переполнение буфера контрольной ленты",
		"",
		"Вносимая клиентом сумма меньше суммы чека",
		"Открыт чек возврата – операция невозможна",
		"Смена превысила 24 часа",
		"Открыт чек продажи – операция невозможна",
		"Переполнение ФП",
		"",
		"Неверный пароль",
		"Буфер контрольной ленты не переполнен",
		"Идет обработка контрольной ленты",
		"Обнуленная касса (повторное гашение невозможно)",
		"",
		"Неверный номер таблицы",
		"Неверный номер ряда",
		"Неверный номер поля",
		"Неверная дата",
		"Неверное время",
		"Сумма чека по секции меньше суммы сторно",
		"Подсчет суммы сдачи невозможен",
		"В ККТ нет денег для выплаты",
		"",
		"Чек закрыт – операция невозможна",
		"Чек открыт – операция невозможна",
		"Смена открыта, операция невозможна",
		"ККТ заблокирована, ждет ввода пароля доступа к ФП",
		"Заводской номер уже задан",
		"Исчерпан лимит перерегистраций",
		"Ошибка ФП",
		"",
		"Неверный номер смены",
		"Неверный тип отчета",
		"Недопустимый пароль",
		"Недопустимый заводской номер ККТ",
		"Недопустимый РНМ",
		"Недопустимый ИНН",
		"ККТ не фискализирована",
		"Не задан заводской номер",
		"Нет отчетов",
		"Режим не активизирован",
		"Нет указанного чека в КЛ",
		"Нет больше записей КЛ",
		"Некорректный код или номер кода защиты ККТ",
		"",
		"Требуется выполнение общего гашения",
		"Команда не разрешена введенными кодами защиты ККТ",
		"Невозможна отмена скидки/надбавки",
		"Невозможно закрыть чек данным типом оплаты (в чеке присутствуют операции без контроля наличных)",
		"Неверный номер маршрута",
		"Неверный номер начальной зоны",
		"Неверный номер конечной зоны",
		"Неверный тип тарифа",
		"Неверный тариф",
		"",
		"Ошибка обмена с фискальным модулем",
		"",
		"",
		"",
		"Необходимо провести профилактические работы",
		"Неверные номера смен в ККТ и ЭКЛЗ",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"Нет устройства, обрабатывающего данную команду",
		"Нет связи с внешним устройством",
		"Ошибочное состояние ТРК",
		"Больше одной регистрации в чеке",
		"Ошибочный номер ТРК",
		"Неверный делитель",
		"",
		"Исчерпан лимит активизаций",
		"Активизация данной ЭКЛЗ в составе данной ККТ невозможна",
		"Перегрев головки принтера",
		"Ошибка обмена с ЭКЛЗ на уровне интерфейса I2C",
		"Ошибка формата передачи ЭКЛЗ",
		"Неверное состояние ЭКЛЗ",
		"Неисправимая ошибка ЭКЛЗ",
		"Авария крипто-процессора ЭКЛЗ",
		"Исчерпан временной ресурс ЭКЛЗ",
		"ЭКЛЗ переполнена",
		"В ЭКЛЗ переданы неверная дата или время",
		"В ЭКЛЗ нет запрошенных данных",
		"Переполнение ЭКЛЗ (итог чека)",
		"Буфер переполнен",
		"Невозможно напечатать вторую фискальную копию",
		"Требуется гашение ЭЖ",
		"Сумма налога больше суммы регистраций по чеку и/или итога или больше суммы регистрации",
		"Начисление налога на последнюю операцию невозможно",
		"Неверный номер ЭКЛЗ",
		"",
		"",
		"Сумма сторно налога больше суммы зарегистрированного налога данного типа",
		"Ошибка SD",
		"Операция невозможна, недостаточно питания"
	];
};
const STXv3 = 0xFE;
const ESCv3 =  0xFD;

kkm.atol.commandv3util = {
	esc: function(data) {
		var i = -1;
		while (++i < data.length)
			if (STXv3 == data[i] || ESCv3 == data[i]) {
				data.splice(i, 1, ESCv3, data[i] - 16);
				i++;
			}
		return data;
	},
	unesc: function(data) {
		var i = 0;
		while(i++<data.length-1)
			if (data[i] == ESCv3)
				data.splice(i, 2, data[i+1] + 16);
		return data;
	},
	crc8table: (function(poly) {
		const crc8val = (c) => { for (var i=0;i<8;++i) { var p = (c & 0x80) ? poly: 0x00; c = ((c << 1) ^ p) % 256; }; return c; };
		const res = [];
		for (var i=0;i<256;++i) res[i] = crc8val(i);
		return res;
	})(0x31),
	crc8: function(data, table) { var res = 0xff; for (var i=0;i<data.length;i++) res = table[(res^data[i]) % 256]; return res;	},
	len:function(l) {
		return [
			l & 0x7f,
			l >> 7
		];
	},
	unlen:function(data) {
		return data[0] + (data[1] << 7);
	},
	pack: function(data, id, esc, crc8, crc8table, len) {
		const lbytes = len(data.length);
		const crc = crc8([id].concat(data), crc8table);
		const escdata = esc(data);

		return [STXv3].
		concat(lbytes).
		concat(id).
		concat(escdata).
		concat(esc([crc]));
	},
	packcompletemaker: function(unlen, unesc, cmdids) {
		return function(b, l) {
			var shft = -1;
			for (var i=0; i<l; i++)
				if (STXv3 == b[i]) shft = i;

			if (-1 == shft) return false;
			if (shft+3 > l) return false;

			const il = unlen([b[shft+1], b[shft+2]]);
			if (!il) return false;
			const data = [];
			for (var i=4; i<l-shft; i++) {
				data[i-4] = b[shft+i];
			}
			if (0xa1 == data[0] || 0xa2 == data[0] || 0xa8 == data[0])
				return false;

			const id = 0xf0 == b[shft+3] ? b[shft+5] : b[shft+3]
			if (!cmdids.includes(id)) return false;

			return (unesc(data).length == il+1);
		};
	},
	newid: function(U, err) {
		return new Promise(function(res) {
			U.fromstorage().
			then(function(items) {
				log("loaded id from storage "+items.id);
				const id = parseInt(items.id || "0x00");
				const newid = id < 0xdf ? id+1: 0x00;
				return U.tostorage({"id": ""+newid}).
				then(function() { return Promise.resolve(newid); });
			}).
			then(function (newid) {
					log("saved to storage id "+newid);
					res(newid);
			}).
			catch(err);
		});
	},
	unpack: function(rawdata, unlen, unesc) {
		const start = rawdata.lastIndexOf(STXv3);
		const data = -1 == start ? rawdata : rawdata.slice(start);
		const l = unlen(data.slice(1, 3));
		const clean = unesc(data.slice(4, data.length - 1));
		if (clean.length < l) throw "packed incomplete";

		return clean;
	},
	tobuf:function(data, tid) {
		return [0xc1,
				0x01,
				tid,
		].concat(data);
	},
	frombuf:function(data, err) {
		const errs = ["buffer full", "duplicate id", "no such id", "illegal value"];
		switch(data[0]) {
			case 0xb1:
			case 0xb2:
			case 0xb3:
			case 0xb4:
				throw "error: "+errs[data[0] - 0xb1];
			case 0xa1:
			case 0xa2:
			case 0xa8:
				throw "command processing incomplete ["+data+"]";
			case 0xa4:
			case 0xa3:
				return data.slice(1);
			case 0xa7:
			case 0xa6:
				return data.slice(2);
			case 0xa5:
				throw "command processing stopped ["+data+"]";
			default:
				throw "receive data unknown";
		}
	},
	cmdids: new Array()
};

kkm.atol54fz.command = function (config, client, logging, conn, send1, receive1, close, flush) {
	const U = kkm.util;
	const L = U.modlog(logging(), "...");
	const V3 = kkm.atol.commandv3util;

	const log = L.log;
	const err = L.err;

	function senddata(id, data) {
		return new Promise(function(res) {
			log("sending command ["+data.map(n=>(n).toString(16))+"] to fr (v3)");
			send1(conn, V3.pack(V3.tobuf(data, id), id, V3.esc, V3.crc8, V3.crc8table, V3.len), function (si) {
				V3.cmdids.push(id);
	            log((!!si && 0 < si.bytesSent && !!si.error) ?
	                "failed sending sending command to fr (v3) "+chrome.runtime.lastError.message :
	                "command ["+data.map(hx)+"] sent to fr (v3)"
	            );
	            res();
			});
		});
	}

	function sendabort(id) {
		return new Promise(function(res) {
			log("flushing all previous commands in command buffer");
			send1(conn, V3.pack([0xc4], id, V3.esc, V3.crc8, V3.crc8table, V3.len), function(si) {
	            log((!!si && 0 < si.bytesSent && !!si.error) ?
	                "failed sending abort command to fr (v3) "+chrome.runtime.lastError.message :
	                "abort command sent to fr (v3)"
	            );
	            res();
			});
		});
	}

	function sendreq(id, cid) {
		return new Promise(function (res) {
			log("rerequesting result for last command sent to FR for id "+id+", cid "+cid+" (v3)");
			send1(conn, V3.pack([0xC3, cid], id, V3.esc, V3.crc8, V3.crc8table, V3.len), function(si) {
				V3.cmdids.push(id);
				log("last command result rerequest sent (v3)");
				log("cmdids buffer is ["+V3.cmdids.map(hx)+"]");
				res();
			});
		});
	}

	function pflush(conn) {
		return new Promise(function(res) {
			flush(conn, res);
		});
	}

	function send (sdata, sentlink) {
		pflush(conn).
		then(function() { return V3.newid(U, err); }).
		then(function (id) { log("sendabort from send"); return sendabort(id); }).
		then(function() { return new Promise(function(res) { setTimeout(res, 200); }); }).
		then(function() { return V3.newid(U, err); }).
		then(function (id) { return senddata(id, sdata); }).
		then(sentlink).
		catch(err);
	};

	function receive(receivedlink) {
		var timer = false;
		function rereq(cd) {
			timer = setTimeout(function() {
				if (6 == cd) return err("requesting command result from FR failed");

				return pflush(conn).
				then(function() { return V3.newid(U, err); }).
				then(function(id) {
					if (!V3.cmdids.length) {
						log("no command to rerequest");
						return U.pjam();
					}

					const cid = V3.cmdids[0];
					log("Rereq started for " + cid + " on attempt " + cd);
					return sendreq(id, cid); }).
				then(function () { rereq(cd + 1); });
			}, 2000);
		}

		log("receiving command response (v3)");
		receive1(conn, V3.packcompletemaker(V3.unlen, V3.unesc, V3.cmdids), function(d) {
			clearTimeout(timer);
			log("cmdids buffer before flushing ["+V3.cmdids.map(hx)+"]")
			try {
				const res = V3.unpack(d, V3.unlen, V3.unesc);
				log("command response received (v3) ["+res.map(hx)+"]");
				const clean = V3.frombuf(res, err);
				log("received response data ["+clean.map(hx)+"] (v3)");
				V3.cmdids.length = 0;
				receivedlink(clean);
			} catch(e) {
				close(function () { err(e); });
			}
		});
		rereq(1);
	}

	client(send, receive, close);
}
kkm.atol54fz.ext = function (logging, config, executor, proc) {
	kkm.atol.ext(logging, config, executor, function (beep, demo, info, type, check, print, z, x, change, cashin, cashout, ofdreport, corrprintpass, hackfr) {
		const U = kkm.util;

		function app(chain) {
			executor(logging, chain);
		}

		function fsapp(chain) {
			executor(U.fslogging(logging, U.id), chain);
		}


		const L = logging();
		const log = L.log;
		const err = L.err;
		const assert = U.assertmaker(config.errtable, err);
	    const assert0 = U.assertmaker(config.errtable, function(msg) { err(msg, 001); });

		function p(cmd) {
			return ((config.password || [0,0]).concat(cmd));
		}

		function talkmaker(send, receive, log) {
			const talk = U.talkmaker(send, receive, log);
			return function(data) {
				return talk(p(data));
			}
		};

		function safetalkmaker(send, receive, close, log, assert) {
			const stalk = U.safetalkmaker(send, receive, close, log, assert);
			return function(data, assertmsg) {
				return stalk(p(data), assertmsg);
			}
		};

		function check54fz(res, checked) {
			check(res, function(res, datas) {
				res.fz54 = true;
				res.vendor = "atol54fz";
				res.qrsupported = true;
				checked(res, datas);
			});
		}

		function print(data, onfinish) {
			data.discount = 0;
			kkm.atol.print.print4pos(config, data, kkm.atol.print.pos54fz, onfinish, app, log, err, assert, talkmaker, safetalkmaker);
		}

		function corrprint(data, onfinish) {
			kkm.atol.print.print4corr(config, data, onfinish, app, log, err, assert, talkmaker, safetalkmaker);
		}

		function reportmaker(cmd, mode, modename) {
			return function(cashier, onfinish) {
				function reportmode(talk, stalk) {
					function modeswitch() {
						log("переход в режим отчетов "+modename)
						return stalk([0x56, mode].concat(config.adminpass), "Ошибка перехода в режим отчетов "+modename);
					}

					function modeexit() {
						log("переход в режим выбора");
						return stalk([0x48], "Ошибка выхода в режим выбора").
						then(function () { return modeswitch(); });
					}

					return talk([0x3f]).then(function(data) {
						const m = data[17];
						log("перед печатью отчета ФР в режиме "+m);

						var proc = modeexit;

						if (0 == m)
							proc = modeswitch;
						if (mode == m)
							proc = U.pass();

						return proc();
					});
				}

				function wait4finish(talk) {
					log("проверяем завершена ли печать отчета")
					function checker() {
						log("отправляем запрос состояния устройства");
						return talk([0x45]).then(function (data) {
							const res = !(0x17 == data[1] || [0x22] == data[1] || [0x23] == data[1]);
							log(res ? "отчет напечатан" : "отчет все еще печатается");
							return res;
						});
					}

					return U.waiter(checker, 500);
				}

				app(
					function(send, receive, close) {
						const talk = talkmaker(send, receive, log);
						const stalk = safetalkmaker(send, receive, close, log, assert);
						const P = kkm.atol.print;
						const setreq = P.setreqmaker(stalk, log);

						reportmode(talk, stalk).
						then(function () {
							return (!!cashier) ?
								setreq(1021, cashier) :
								Promise.resolve();
						}).
						then(function () { return stalk(cmd, "Ошибка запроса отчета"); }).
						then(function () { return wait4finish(talk); }).
						then(function () { close(onfinish); });
					}
				);
			}
		}

		function moderegister(stalk, talk) {
			return talk([0x3f]).
			then(function (data) {
				function reg() {
					return stalk([0x56, 0x01].concat(config.adminpass), "Ошибка при переходе в режим регистрации");
				}

				if (0 != (data[9] & 2)) {
					return Promise.reject(new Error("Смена уже открыта"));
				}

				if (1 == data[17])
					return Promise.resolve();

				if (0 == data[17])
					return reg();

				return stalk([0x48], "Ошибка при переходе в режим выбора").then(reg);
			});
		}

		function changeext() {
			const l = arguments.length;
			const cb = arguments[l-1];
			const cashier = 1<l ? arguments[0] : false;

			app(function(send, receive, close) {
				const stalk = safetalkmaker(send, receive, close, log, assert0);
				const talk = talkmaker(send, receive, log);
				const P = kkm.atol.print;
				const setreq = P.setreqmaker(stalk, log);

				return moderegister(stalk, talk).
				then(function () {
					return (!!cashier) ?
						setreq(1021, cashier) :
						Promise.resolve();
				}).
				then(function() {
					return stalk([0x9a, 0x00], "Ошибка при открытии смены");
				}).
				then(function() { close(cb); }).
				catch(function(e) { close(function() { err(e.message ? e.message : e); }); });
			});
		}

		function hackfr(data, onfinish) {
			const AP = kkm.atol.print;

			fsapp(function(send, receive, close) {
				const talk = talkmaker(send, receive, log);
				const fsassert = U.assertmaker(config.errtable, log);
				const fstalk = safetalkmaker(send, receive, close, log, fsassert);

				const setfield = AP.setfieldmaker(log, log, talk, fstalk, config);

				setfield(2, 1, 115, [0x01]).
				then(function () {
					close(onfinish);
				}).
				catch(function(e) {
					close(function() { log(e.message || e); });
				}) ;
			});
		}

		const zext = reportmaker([0x5a], 0x03, "с гашением");
		const xext = reportmaker([0x67,0x01], 0x02, "без гашения");
		const ofdext = reportmaker([0x67,0x09], 0x02, "без гашения");
		proc(beep, demo, info, type, check54fz, print, zext, xext, changeext, cashin, cashout, ofdext, corrprint, hackfr);
	});
}
kkm.atol54fz.errors = function () {	return ["Ошибок нет","Контрольная лента обработана без ошибок",
"",
"",
"",
"",
"",
"",
"Неверная цена (сумма)",
"",
"Неверное количество",
"Переполнение счетчика наличности",
"Невозможно сторно последней операции",
"Сторно по коду невозможно (в чеке зарегистрировано меньшее количество товаров с указанным кодом)",
"Невозможен повтор последней операции",
"Повторная скидка на операцию невозможна",
"Скидка/надбавка на предыдущую операцию невозможна",
"Неверный код товара",
"Неверный штрихкод товара",
"Неверный формат",
"Неверная длина",
"ККТ заблокирована в режиме ввода даты",
"Требуется подтверждение ввода даты",
"",
"Нет больше данных для передачи ПО ККТ",
"Нет подтверждения или отмены продажи",
"Отчет с гашением прерван. Вход в режим невозможен.",
"Отключение контроля наличности невозможно (не настроены необходимые типы оплаты).",
"",
"",
"Вход в режим заблокирован",
"Проверьте дату и время",
"Дата и время в ККТ меньше чем в ФН",
"Невозможно закрыть архив",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"Товар не найден",
"Весовой штрихкод с количеством <>1.000",
"Переполнение буфера чека",
"Недостаточное количество товара",
"Сторнируемое количество больше проданного",
"Заблокированный товар не найден в буфере чека",
"Данный товар не продавался в чеке, сторно невозможно",
"",
"",
"Неверная команда от ККТ",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"",
"Команда не реализуется в данном режиме ККТ",
"Нет бумаги",
"Нет связи с принтером чеков",
"Механическая ошибка печатающего устройства",
"Неверный тип чека",
"Нет больше строк картинки/штрихкода",
"Неверный номер регистра",
"Недопустимое целевое устройство",
"Нет места в массиве картинок/штрихкодов",
"Неверный номер картинки/штрихкода (картинка/штрихкод отсутствует)",
"Сумма сторно больше, чем было получено данным типом оплаты",
"Сумма не наличных платежей превышает сумму чека",
"Сумма платежей меньше суммы чека",
"Накопление меньше суммы возврата или аннулирования",
"",
"Переполнение суммы платежей",
"Предыдущая операция незавершена",
"Ошибка GSM-модуля",
"",
"",
"Данная модель ККТ не может выполнить команду",
"Неверная величина скидки / надбавки",
"Операция после скидки / надбавки невозможна",
"Неверная секция",
"Неверный вид оплаты",
"Переполнение при умножении",
"Операция запрещена в таблице настроек",
"Переполнение итога чека",
"Открыт чек аннулирования – операция невозможна",
"",
"Переполнение буфера контрольной ленты",
"",
"Вносимая клиентом сумма меньше суммы чека",
"Открыт чек возврата – операция невозможна",
"Смена превысила 24 часа",
"Открыт чек продажи – операция невозможна",
"Переполнение ФП",
"",
"Неверный пароль",
"Буфер контрольной ленты не переполнен",
"Идет обработка контрольной ленты",
"Обнуленная касса (повторное гашение невозможно)",
"",
"Неверный номер таблицы",
"Неверный номер ряда",
"Неверный номер поля",
"Неверная дата",
"Неверное время",
"Сумма чека по секции меньше суммы сторно",
"Подсчет суммы сдачи невозможен",
"В ККТ нет денег для выплаты",
"",
"Чек закрыт – операция невозможна",
"Чек открыт – операция невозможна",
"Смена открыта, операция невозможна",
"",
"Заводской номер уже задан",
"",
"",
"",
"Неверный номер смены",
"Неверный тип отчета",
"Недопустимый пароль",
"Недопустимый заводской номер ККТ",
"Недопустимый РНМ",
"Недопустимый ИНН",
"ККТ не фискализирована",
"Не задан заводской номер",
"Нет отчетов",
"Режим не активизирован",
"Нет указанного чека в КЛ",
"Нет больше записей КЛ",
"Некорректный код или номер кода защиты ККТ",
"Отсутствуют данные в буфере ККТ",
"Требуется выполнение общего гашения",
"Команда не разрешена введенными кодами защиты ККТ",
"Невозможна отмена скидки/надбавки",
"Невозможно закрыть чек данным типом оплаты (в чеке присутствуют операции без контроля наличных)",
"Неверный номер маршрута",
"Неверный номер начальной зоны",
"Неверный номер конечной зоны",
"Неверный тип тарифа",
"Неверный тариф",
"",
"Ошибка обмена с фискальным модулем",
"",
"",
"",
"Необходимо провести профилактические работы",
"Неверные номера смен в ККТ и ФН",
"",
"",
"",
"",
"",
"",
"",
"",
"Нет устройства, обрабатывающего данную команду",
"Нет связи с внешним устройством",
"Ошибочное состояние ТРК",
"Больше одной регистрации в чеке",
"Ошибочный номер ТРК",
"Неверный делитель",
"",
"",
"Активизация данного ФН в составе данной ККТ невозможна",
"Перегрев головки принтера",
"Ошибка обмена с ФН на уровне интерфейса I2C",
"Ошибка формата передачи ФН",
"Неверное состояние ФН",
"Неисправимая ошибка ФН",
"Ошибка КС ФН",
"Закончен срок эксплуатации ФН",
"Архив ФН переполнен",
"В ФН переданы неверная дата или время",
"В ФН нет запрошенных данных",
"Переполнение ФН (итог чека)",
"Буфер переполнен",
"Невозможно напечатать вторую фискальную копию",
"Требуется гашение ЭЖ",
"Сумма налога больше суммы регистраций по чеку и/или итога или больше суммы регистрации",
"Начисление налога на последнюю операцию невозможно",
"Неверный номер ФН",
"",
"",
"Сумма сторно налога больше суммы зарегистрированного налога данного типа",
"Ошибка SD",
"Операция невозможна, недостаточно питания",
"Некорректное значение параметров команды ФН",
"Превышение размеров TLV данных ФН",
"Нет транспортного соединения ФН",
"Исчерпан ресурс КС ФН",
"Исчерпан ресурс хранения ФН",
"Сообщение от ОФД не может быть принято ФН",
"В ФН есть неотправленные ФД",
"Ошибка ККТ. Убедитесь что ККТ зарегистрирована только для одной СНО. Убедитесь что если ККТ зарегистрирована как интернет касса, в чек передаются адрес или телефон покупателя."];};
kkm.shtrikh.command = function (config, client, logging, conn, send1, receive1, close, flush, state) {
	const U = kkm.util;

	const crc = U.crc;
	const bytechecker = U.bytecheck;

	const L = U.modlog(logging(), "......");

	const log = L.log;
	const err = L.err;

	function pack(data) {
		data.unshift(data.length);
		data.push(crc(data));
		data.unshift(STX);

		log("packed for sending "+data.map(hx));
		return data;
	}

	var lastcommand = 0;

	function packet(b, l) {
		function part(b, from, to) {
			const res = [];
			for (var i=from; i<to; i++)
				res.push(b[i]);
			return res;
		}

		for (var i=l-3; i>=0; i--) {
			if (STX == b[i]) {
				const blen = b[i+1]
				if (blen && l-i >= blen+3) {
					if (b[i+2] != lastcommand) continue;
					const bcrc = b[i+blen+2];
					const bdata = part(b, i+1, i+blen+2);
					if (bcrc == crc(bdata))
						return part(b, i, i+blen+3);
				}
			}
		}

		return false;
	}

	function unpack(data) {
		return packet(data, data.length).slice(2, -1);
	}

	var lastsend = function () { log("last send can not be repeated, nothing been sent yet"); };
	function send (data, sentlink) {
		var attempt = 0;
		(lastsend = function () {
			if (++attempt == config.retries) {
				log("send failed in "+config.retries+" attempts");
				close();
				return;
			}

			log("sending command "+data+", attempt "+attempt);
			lastcommand = data[0];
			send1(conn, pack(data), function() {
				if (chrome.runtime.lastError) {
					log("failed sending data, "+chrome.runtime.lastError.message);
					setTimeout(lastsend, config.porttimeout);
				} else {
					log("data sent!");
					sentlink(send, receive);
				}
			});
		})();
	};

	function receive(receivedlink) {
		var receivetimeout = false;
		(function retry(attempt) {
			if (config.retries == attempt) {
				log("receive failed in "+config.retries+" attempts");
				return;
			}

			log("receiveing response, attempt "+attempt+" of "+config.retries);
			receivetimeout = setTimeout(function () {
				log("receive timed out");
				retry(attempt+1);
				send1(conn, [NAK, ENQ], function () {
					log("rerequesting receive");
				});
			}, config.porttimeout);
		})(0);

		receive1(conn, packet, function (d) {
			if (receivetimeout)
				clearTimeout(receivetimeout);

			const data = unpack(d);
			const succ = !!data;
			send1(conn, (succ ? [ACK] : [NAK]), function () {
				log((succ ? "confirmed" : "declined")+" data receive");
				if (succ) flush(conn, () => receivedlink(data, send, receive));
			});
		});
	}

	function initcheck() {
		function confirmprev() {
			log("confirming previous command");
			receive1(conn, function(d, l) { return !!packet(d, l); }, function(d) {
				log("previous command confirmed, processing")
				client(send, receive, close, state);
			});
			send1(conn, [ACK], function() {});
		}

		receive1(conn, bytechecker, function (d) {
			log("on initial received ["+d.map(hx)+"]");
			switch (d[0]) {
				case NAK:
					log("init successfull");
					client(send, receive, close, state);
					break;
				case ACK:
					log("need to confirm prev command");
					confirmprev();
					break;
				default:
					send1(conn, [ACK], function () {
						setTimeout(function (){
							initcheck();
						}, config.porttimeout);
					});
			}
		});
		log("sending initial check");
		send1(conn, [ACK,ENQ], function (si) {
			if (si.error) return err("Ошибка инициализации устройства(возможно некорректная скорость COM-порта): "+si.error);
			log("initial check sent");
		});
	}

	initcheck();
}
kkm.shtrikh54fz.command = kkm.shtrikh.command;
kkm.shtrikh.api = function (logging, config, executor, proc) {
	const U = kkm.util;

	const L = U.modlog(logging(), "...");

	const log = L.log;
	const err = L.err;

	const assert = U.assertmaker(config.errtable, err);
    const assert0 = U.assertmaker(config.errtable, function(msg) { err(msg, 001); });

	function app(chain) {
		executor(logging, chain);
	}

	function beep(data, onfinish) {
		app(
			function (send, receive, close) {
				receive(assert("Ошибка при звуковом сигнале", close, function (data) {
					log("beep confirm sais: "+(0 == data[1] ? "beeeeep!" : "error "+data[1]));
					close();
				}));
				send([0x13].concat(config.operpass), function() {
					log("beep sent");
				});
			}
		);
	}

	function demo(data, onfinish) {
		app(
			function (send, receive, close) {
				receive(assert("Ошибка при тестовом прогоне", close, function (data) {
					close();
					if (onfinish && onfinish instanceof Function)
						onfinish();
				}));
				send([0x19].concat(config.operpass).concat([0x01]), function () {
					log("test pass request sent");
				});
			}
		);
	}

function demotext(data, onfinish) {
		app(
			function (send, receive, close) {
				const stalk = U.safetalkmaker(send, receive, close, log, assert);
				const mas = [kkm.config.default.demotext,"","","","","","","","",""];
				const text = mas.map(s => cutbytes(win1251encode(s),20));
				function printline(prev, line) {
					return prev.then(function () {
						return stalk([0x17].concat(config.operpass).concat([3]).concat(cutbytes(line, 40)), "Ошибка при печати строки (" + line + ")");
					});
				}
				text.reduce(printline, Promise.resolve()).then(
					function ()
					{
						close();
						if (onfinish && onfinish instanceof Function)
							onfinish();
					}
				);
			}
		);
	}

	function info(proc) {
		app(
			function(send, receive, close) {
				receive(assert("Ошибка при получении информации об устройстве", close, function (data) {
					close(function () { proc(data); });
				}));
				log("запрос состояния ФР Штрих");
				send([0x11].concat(config.operpass), function() {
					log("long info request sent");
				});
			}
		);
	}

	function type(proc) {
		app(
			function(send, receive, close) {
				receive(assert("Ошибка при получении типа устройства", close, function (data) {
					log("type data received");
					close(function () { proc(data); });
				}));
				send([0xfc], function() {
					log("type request sent");
				});
			}
		);
	}

	function check(type, info) {
		app(
			function (send, receive, close) {
				function t(f) {
					receive(function (data) {
						log(data[1] ? "error "+conf.errtable[data[1]] : "type data received");
						kkm.shtrikh.typedata = data;
						f();
					});
					send([0xfc], function() {
						log("type request sent!");
					});
				}

				function i() {
					type(kkm.shtrikh.typedata);
					receive(function (data) {
                        log("info data received!");
                        close(function () { info(data); });
					});
					send([0x11].concat(config.operpass), function() {
						log("info request sent!");
					});
				}
				(!!kkm.shtrikh.typedata ? U.empty : t)(i);
			}
		);
	}

	function wait4finish(talk) {
		log("ожидаем окончания процесса");
		return U.waiter(function() {
			return talk([0x10].concat(config.operpass)).
			then(function (infodata) { return new Promise(function (r) {
				const ready = (5 < infodata.length && 0 == infodata[6]);
				r(ready);
			}); });
		}, 2000);
	}

	function cutreceipt(stalk) {
		return  stalk([0xfc], "Ошибка при получении типа устройства").
				then(function(data) {
					const model = data[6];
					return config.nocutmodels.includes(model) ?
						Promise.resolve() :
						stalk([0x25].concat(config.operpass).concat([1]), "Ошибка отрезки чека");
				});
	}

	function print(receipt, onfinish) {
		const linelengths = {
			250:57,
			7:50,
			4:36,
			252:32,
			240:32,

		};

		var model;
		function check(data, fail, drop) {
			const modes = {
				1: "Выдача данных",
				5: "Блокировка по неправильному паролю налогового инспектора",
				6: "Ожидание подтверждения ввода даты",
				7: "Разрешение изменения положения десятичной точки",
				9: "Режим разрешения технологического обнуления",
				10: "Тестовый прогон.",
				11: "Печать полного фис. отчета",
				12: "Печать отчёта ЭКЛЗ",
				13: "Работа с фискальным подкладным документом",
				14: "Печать подкладного документа"
			};

			switch (true) {
				case (8 == data[5]):
					return drop();
				case (3 == data[5]):
					return fail("Ошибка при печати кассового чека. Смена превысила 24 часа, для продолжения работы необходимо снять Z-отчет");
				case (0 == data[5] || 4 == data[5] || 2 == data[5] || 15 == data[5]):
					return Promise.resolve();
				default: return fail("Печать чека невозможна в режиме: "+(modes[data[5]] || +data[5]));
			}
		}

		const ll = function () { return (linelengths[model] || 40); };

		const ofdblock = U.ofd(
			win1251encode,
			fillarr(0x2d, ll()),
			[0x2a, 0x2a, 0x2a, 0x20, 0xd4, 0xd0, 0x20, 0x2a, 0x2a, 0x2a],
			[0x3a, 0x20]
		)

		log("печатаем чек "+(receipt.return?"возврата":"продажи"));
		const rtype = receipt.return ? [0x02] : [0x00];
		const stype = receipt.return ? [0x82] : [0x80];

		app(
			function (send, receive, close, state) {
				const talk = U.talkmaker(send, receive, log);
				const stalk = U.safetalkmaker(send, receive, close, log, assert);
				const stalk0 = U.safetalkmaker(send, receive, close, log, assert0);

				function printline(prev, line) {
					return prev.then(function () {
						return stalk([0x17].concat(config.operpass).concat([3]).concat(cutbytes(line, 40)), "Ошибка при печати строки (" + line + ")");
					});
				}

				function sell(prev, rl) {
					log("печать строки "+JSON.stringify(rl));

					const quantity = int2bytes(Math.round(rl.quantity*1000), [0,0,0,0,0]).reverse();
					const amount = int2bytes(rl.price, [0,0,0,0,0]).reverse();
					const name = splitbytes(win1251encode(rl.name), ll());

					log("данные подготовлены, amount "+amount.join()+", quantity: "+quantity.join()+", name: "+name.join());

					return prev.then(function () {
						return name.reduce(printline, Promise.resolve()).
						then(function () { return stalk(stype.concat(config.operpass).concat(quantity).concat(amount).concat(zeroarr(45)), "Ошибка печати строки чека") });
					});
				}

				function fillmodel(withmodel) {
					return stalk([0xfc], "Ошибка при получении типа устройства").
					then(function (data) {
						model = data[6];
						log("получили модель Штриха ("+model+")");
						return Promise.resolve();
					});
				}

				function drop(succ) {
					return stalk([0x88].concat(config.operpass), "Ошибка аннулирования незакрытого чека").
					then(function () {
						log("предыдущий не закрытый чек аннулирован");
						return new Promise(function(r) {
							setTimeout(r, 200);
						});
					});
				}

				function fail(msg) {
					return new Promise(function(res) {
						close(function () {
							err(msg);
							res();
						});
					});
				}

				function upload(data) {
					function upload(data, block) {
						log("Uploading " + data + " into block " + block);
						return stalk(
							[0xdd].
							concat(config.operpass).
							concat([0x00]).
							concat(block).
							concat(data)
							, "Ошибка загрузки данных")
					}

					const blocks = splitbytes(data, 64);
					return blocks.reduce(
						(p, block, index) => {
							log("Uploading " + JSON.stringify(block) + " into block " + index);
							return p.then(() => upload(block, index));
						}, Promise.resolve());
				}

				function egaisprint(receipt) {
					function qr() {
						if (kkm.shtrikh.qr_supported) {
                            log("Printing QR code");
                            const data = win1251encode(receipt.url);
                            return upload(data).then(() => stalk(
                                [0xde].
                                concat(config.operpass).
                                concat([
                                    0x03,
                                    data.length,
                                    0x00,
                                    0x00,
                                    0x00,
                                    0x00,
                                    0x06,
                                    0x00,
                                    0x02,
                                    0x01
                                ])
                            ));
                        } else {
							log("Skipping QR code print");
							return Promise.resolve();
						}
					}

					const head = kkm.egais.head(win1251encode, receipt, kkm.shtrikh.cashbox, ll());
					const tail = kkm.egais.tail(win1251encode, receipt, ll());

					return head.reduce(printline, Promise.resolve())
						.then(qr)
						.then(() => tail.reduce(printline, Promise.resolve()));
				}

				fillmodel().
				then(function () { return talk([0x10].concat(config.operpass)); }).
				then(function (data) { return check(data, fail, drop); }).
				then(function () { return stalk0([0x8d].concat(config.operpass).concat(rtype), "Ошибка открытия чека")}).
				then(function () {
					log("чек открыт");
					const block = ofdblock(receipt.ofdCode, receipt.ofdKKT, receipt.ofdTime, false);

					return block.reduce(printline, Promise.resolve());
				}).
				then(function () { return receipt.positions.reduce(sell, Promise.resolve()); }).
				then(function () {
					return receipt.discount ?
						stalk([0x86].concat(config.operpass).concat(int2bytes(receipt.discount, [0,0,0,0,0]).reverse()).concat(zeroarr(44)), "Ошибка регистрации скидки"):
						Promise.resolve();
				}).
				then(function () { return stalk([0x89].concat(config.operpass), "Ошибка печати подытога чека"); }).
				then(function () {
					const cash = int2bytes(receipt.paidSum, [0,0,0,0,0]).reverse();
					const nocash = int2bytes(receipt.noCashSum, [0,0,0,0,0]).reverse();
					const summ =  cash.concat(zeroarr(10)).concat(nocash.concat(zeroarr(46)));

					return stalk([0x85].concat(config.operpass).concat(summ), "Ошибка закрытия чека");
				}).
				then(function() { state.stagecritical = false; }).
				then(function () { return wait4finish(talk).then(function () { return new Promise(function (r) { setTimeout(r, 200); }) }); }).
				then(() => receipt.egais1 ? egaisprint(receipt.egais1) : Promise.resolve()).
				then(function () { return cutreceipt(stalk) }).
				then(function () { return new Promise(function(r) { close(r); }); }).
				then(onfinish);
			}
		);
	}

	function z(data, onfinish) {
		app(function(send, receive, close) {
			const talk = U.talkmaker(send, receive, log);
			const stalk = U.safetalkmaker(send, receive, close, log, assert);

			stalk([0x41].concat(config.adminpass), "Ошибка печати Z-отчета").
			then(function () { return wait4finish(talk).then(function () { return new Promise(function (r) { setTimeout(r, 200); }) }); }).
			then(function () { return new Promise(function(r) { close(r); }); }).
			then(onfinish);
		});
	}

	function x(data, onfinish) {
		app(function(send, receive, close) {
			const talk = U.talkmaker(send, receive, log);
			const stalk = U.safetalkmaker(send, receive, close, log, assert);

			stalk([0x40].concat(config.adminpass), "Ошибка печати X-отчета").
			then(function () { return wait4finish(talk).then(function () { return new Promise(function (r) { setTimeout(r, 200); }) }); }).
			then(function () { return new Promise(function(r) { close(r); }); }).
			then(onfinish);
		});
	}

	function change(data, onfinish) {
		if (onfinish && onfinish instanceof Function) onfinish();
	}

	function cashin(data, onfinish) {
		app(function(send, receive, close) {
			const summ = data.sum || data;
			const talk = U.talkmaker(send, receive, log);
			const stalk = U.safetalkmaker(send, receive, close, log, assert);

			stalk([0x50].concat(config.operpass).concat(int2bytes(summ, [0,0,0,0,0]).reverse()), "Ошибка при внесении наличных").
			then(function() { return wait4finish(talk) }).
			then(function () { return cutreceipt(stalk); }).
			then(function () { return new Promise(function(r) { close(r); }); }).
			then(onfinish);
		});
	}

	function cashout(data, onfinish) {
		app(function(send, receive, close) {
			const summ = data.sum || data;
			const talk = U.talkmaker(send, receive, log);
			const stalk = U.safetalkmaker(send, receive, close, log, assert);

			stalk([0x51].concat(config.operpass).concat(int2bytes(summ, [0,0,0,0,0]).reverse()), "Ошибка при выплате наличных").
			then(function() { return wait4finish(talk); }).
			then(function () { return cutreceipt(stalk); }).
			then(function () { return new Promise(function(r) { close(r); }); }).
			then(onfinish);
		});
	}

	proc(beep, /*demo*/demotext, info, type, check, print, z, x, change, cashin, cashout, wait4finish);
}
kkm.shtrikh.ext = function (logging, config, executor, proc) {
	kkm.shtrikh.api(logging, config, executor, function (beep, demo, info, type, check, print, z, x, change, cashin, cashout, wait4finish) {
		const U = kkm.util;

		const L = U.modlog(logging(), "...");

		const log = L.log;
		const err = L.err;

		function app(chain) {
			executor(logging, chain);
		}

		const checkext = function (res, checked) {
			const datas = [];
			app(
				function (send, receive, close) {
					function t() {
						receive(function (data) {
							log(data[1] ? "error "+conf.errtable[data[1]] : "type data received");
							datas.push(data);
							res.model = win1251decode(data.slice(8, data.length));
							i();
						});
						send([0xfc], function() {
							log("type request sent!");
						});
					}

					function i() {
						receive(function (data) {
							log("info data received!");
  						kkm.shtrikh.cashbox = data[10];
              log("Cashbox is: " + kkm.shtrikh.cashbox);
							datas.push(data);
							if(data[3] && data[4]) res.ver = " v. "+win1251decode([data[3], data[4]]).split("").join(".");
							else res.ver = "";
							res.fiscal = Boolean(data[2] & 32);
				      res.doc = bytes2int([0, 0, data[11], data[12]]);
				      res.receipt = 0;
				      res.change = bytes2int([0, 0, data[36], data[37]])+1;
				      res.serial = ""+bytes2int([data[35], data[34], data[33], data[32]]);
							q();
						});
						send([0x11].concat(config.operpass), function() {
							log("info request sent!");
						});
					}

					function q() {
						receive(data => {
                            res.qrsupported = (data[1] == 0x33);
							log("QR code test ended, result: " + res.qrsupported);
                            fz();
						});

						send([0xde].concat(config.operpass).concat([
                            0x03,
                            0x00,
                            0x00,
                            0x00,
                            0x00,
                            0x00,
                            0x00,
                            0x00,
                            0x02,
                            0x01
						]), () => log("QR code test started"));
					}

					function fz() {
						receive((data) => {
							res.fz54 = data.length > 2 && (0xff == data[0] && 0x01 == data[1]);
							res.vendor = res.fz54 ? "shtrikh54fz" : "shtrikh";
							datas.push(data);
							log("Получен ответ на запрос статуса ФН, ФР Штрих "+(res.fz54?"":"не ")+"поддерживает ФЗ-54");
							close(function () { checked(res, datas); });
						});
						send([0xff, 0x01].concat(config.adminpass), function() {
							log("Запрос статуса ФН отправлен");
						})
					}

					t();
				}
			);
		}

		proc(beep, demo, info, type, checkext, print, z, x, change, cashin, cashout, wait4finish, /*ofdreport*/U.cbpass, /*corrprint*/U.cbpass, /*hackfr*/U.cbpass);
	});
}
kkm.shtrikh54fz.ext = function (logging, config, executor, proc) {
	kkm.shtrikh.ext(logging, config, executor, function (beep, demo, info, type, check, print, z, x, change, cashin, cashout, wait4finish, ofdreport, corrprintpass, hackfr) {
		const U = kkm.util;

		const L = U.modlog(logging(), "...");

		const log = L.log;
		const err = L.err;

		const assert = U.assertmaker(config.errtable, err);

		const X = kkm.ext;

		function app(chain) {
			executor(logging, chain);
		}

		function cutreceipt(stalk) {
			return  stalk([0xfc], "Ошибка при получении типа устройства").
					then(function(data) {
						const model = data[6];
						return config.nocutmodels.includes(model) ?
							Promise.resolve() :
							stalk([0x25].concat(config.operpass).concat([1]), "Ошибка отрезки чека");
					});
		}

		function e0(code) {
		    const msg = config.errtable[code] || "Ошибка при открытии смены ("+code+")";
		    const e = new Error(msg);
		    e.code = 001;
		    return e;
		}

		function corrprint(corrreceipt, onfinish)
		{
			var model;
			function check(data, fail, drop, open) {
				const modes = {
					1: "Выдача данных",
					5: "Блокировка по неправильному паролю налогового инспектора",
					6: "Ожидание подтверждения ввода даты",
					7: "Разрешение изменения положения десятичной точки",
					9: "Режим разрешения технологического обнуления",
					10: "Тестовый прогон.",
					11: "Печать полного фис. отчета",
					12: "Печать отчёта ЭКЛЗ",
					13: "Работа с фискальным подкладным документом",
					14: "Печать подкладного документа"
				};

				switch (true) {
					case (4 == data[5]):
						return open();
					case (8 == data[5] || 40 == data[5]):
						return drop();
					case (3 == data[5]):
						return fail("Ошибка при печати кассового чека. Смена превысила 24 часа, для продолжения работы необходимо снять Z-отчет");
					case (0 == data[5] || 2 == data[5] || 15 == data[5]):
						return Promise.resolve();
					default: return Promise.reject("Печать чека невозможна в режиме(подрежиме): "+(modes[data[5]] || data[5])+"("+data[6]+")");
				}
			}

			app(
				function (send, receive, close) {
					const talk = U.talkmaker(send, receive, log);
					const stalk = U.safetalkmaker(send, receive, close, log, assert);

					const setreq = setreqmaker(stalk);


					function fillmodel(withmodel) {
						log("получаем модель");
						return stalk([0xfc], "Ошибка при получении типа устройства").
						then(function (data) {
							model = data[6];
							log("получили модель Штриха ("+model+")");
							return Promise.resolve();
						});
					}

					function open() {
						return talk([0xE0].concat(config.operpass)).
						then(function (data) {
						    if (data[1]) throw e0(data[1]);
						    else {
							    log("ждем когда смена откроется");
							    return wait4finish(talk);
							}
						}).then(function () {
							log("Смена открыта");
							return Promise.resolve();
						}).
						catch (function(e) {
						    close(function() { err(e.message || e, e.code); });
						});
					}

					function drop() {
						return stalk([0x88].concat(config.operpass), "Ошибка аннулирования незакрытого чека").
						then(function () {
							log("предыдущий не закрытый чек аннулирован");
							return new Promise(function(r) {
								setTimeout(r, 200);
							});
						});
					}

					function fail(msg) {
						return Promise.reject(new Error(msg));;
					}

					fillmodel().
					then(function () {
						log("получаем состояние ФР");
						return talk([0x10].concat(config.operpass));
					}).
					then(function (data) {
						log("проверяем готовность ФР")
						return check(data, fail, drop, open);
					}).
					then(setcashiermaker(corrreceipt.cashier, stalk, log, 30)).
					then(function () {
						log("печатаем чек коррекции");
						const cashsum = int2bytes(corrreceipt.cashsum, [0,0,0,0,0]).reverse();
						const nocashsum = int2bytes(corrreceipt.nocashsum, [0,0,0,0,0]).reverse();
						const sum = int2bytes(corrreceipt.nocashsum+corrreceipt.cashsum, [0,0,0,0,0]).reverse();
						return stalk([0xff,0x4a]
						.concat(config.adminpass).concat(corrreceipt.type).concat(corrreceipt.op)
						.concat(sum).concat(cashsum).concat(nocashsum).concat(zeroarr(46)),
						"Ошибка при печати чека коррекции");
					}).
					then(function () {
						log("отключаемся");
						return new Promise(function(r) { close(r); });
					}).
					then(onfinish).
                    catch(e => {
                        close(function () { err(e.message || e); });
                    });
				}
			);
		}

		function printext(receipt, onfinish) {
			const linelengths = {
				250:57,
				7:50,
				4:36,
				252:32,
				240:32,
			};

			const vats = {
				18.0: 1,
				10.0: 2,
				0: 3
			};

			var model;
			function check(data, fail, drop, open) {
				const modes = {
					1: "Выдача данных",
					5: "Блокировка по неправильному паролю налогового инспектора",
					6: "Ожидание подтверждения ввода даты",
					7: "Разрешение изменения положения десятичной точки",
					9: "Режим разрешения технологического обнуления",
					10: "Тестовый прогон.",
					11: "Печать полного фис. отчета",
					12: "Печать отчёта ЭКЛЗ",
					13: "Работа с фискальным подкладным документом",
					14: "Печать подкладного документа"
				};

				switch (true) {
					case (4 == data[5]):
						return open();
					case (8 == data[5] || 40 == data[5]):
						return drop();
					case (3 == data[5]):
						return fail("Ошибка при печати кассового чека. Смена превысила 24 часа, для продолжения работы необходимо снять Z-отчет");
					case (0 == data[5] || 2 == data[5] || 15 == data[5]):
						return Promise.resolve();
					default: return Promise.reject("Печать чека невозможна в режиме(подрежиме): "+(modes[data[5]] || data[5])+"("+data[6]+")");
				}
			}

			const ll = function () { return (linelengths[model] || 40); };

			const ofdblock = U.ofd(
				win1251encode,
				fillarr(0x2d, ll()),
				[0x2a, 0x2a, 0x2a, 0x20, 0xd4, 0xd0, 0x20, 0x2a, 0x2a, 0x2a],
				[0x3a, 0x20]
			)

			log("печатаем чек "+(receipt.return?"возврата":"продажи"));
			const rtype = receipt.return ? [0x02] : [0x00];
			const stype = receipt.return ? [0x82] : [0x80];

			app(
				function (send, receive, close, state) {
					const talk = U.talkmaker(send, receive, log);
					const stalk = U.safetalkmaker(send, receive, close, log, assert);

					const setreq = setreqmaker(stalk);

					function printline(prev, line) {
						return prev.then(function () {
							return stalk([0x17].concat(config.operpass).concat([3]).concat(cutbytes(line, 40)), "Ошибка при печати строки (" + line + ")");
						});
					}

					function sell(prev, rl) {
						return prev.
							then(function () {
								log("печать строки "+JSON.stringify(rl));

								const quantity = int2bytes(Math.round(rl.quantity*1000), [0,0,0,0,0]).reverse();
								const amount = int2bytes(rl.price, [0,0,0,0,0]).reverse();
								const name = cutbytes(win1251encode(rl.name), 40);

								log("данные подготовлены, amount "+amount.map(hx).join()+", quantity: "+quantity.map(hx).join()+", name: "+name.map(hx).join());

								const p = stype.concat(config.operpass).concat(quantity).concat(amount).concat([0, (vats[rl.vat] || 4), 0, 0, 0]).concat(name);
								return stalk(p, "Ошибка печати строки чека");
							});
					}

					function fillmodel(withmodel) {
						log("получаем модель");
						return stalk([0xfc], "Ошибка при получении типа устройства").
						then(function (data) {
							model = data[6];
							log("получили модель Штриха ("+model+")");
							return Promise.resolve();
						});
					}

					function open() {
						return talk([0xE0].concat(config.operpass)).
						then(function (data) {
						    if (data[1]) throw e0(data[1]);
						    else {
							    log("ждем когда смена откроется");
							    return wait4finish(talk);
							}
						}).then(function () {
							log("Смена открыта");
							return Promise.resolve();
						}).
						catch (function(e) {
						    close(function() { err(e.message || e, e.code); });
						});
					}

					function drop() {
						return stalk([0x88].concat(config.operpass), "Ошибка аннулирования незакрытого чека").
						then(function () {
							log("предыдущий не закрытый чек аннулирован");
							return new Promise(function(r) {
								setTimeout(r, 200);
							});
						});
					}

					function fail(msg) {
						return Promise.reject(new Error(msg));;
					}

					function upload(data) {
						function upload(data, block) {
							log("Uploading " + (data.map(hx)) + " into block " + block);
							return stalk(
								[0xdd].
								concat(config.operpass).
								concat([0x00]).
								concat(block).
								concat(data)
								, "Ошибка загрузки данных")
						}

						const blocks = splitbytes(data, 64);
						return blocks.reduce(
							(p, block, index) => {
								log("Uploading " + JSON.stringify(block) + " into block " + index);
								return p.then(() => upload(block, index));
							}, Promise.resolve());
					}

					function egaisprint(receipt) {
						function qr() {
							if (kkm.shtrikh.qr_supported) {
	                            log("Printing QR code");
	                            const data = win1251encode(receipt.url);
	                            return upload(data).then(() => stalk(
	                                [0xde].
	                                concat(config.operpass).
	                                concat([
	                                    0x03,
	                                    data.length,
	                                    0x00,
	                                    0x00,
	                                    0x00,
	                                    0x00,
	                                    0x06,
	                                    0x00,
	                                    0x02,
	                                    0x01
	                                ])
	                            ));
	                        } else {
								log("Skipping QR code print");
								return Promise.resolve();
							}
						}

						const head = kkm.egais.head(win1251encode, receipt, kkm.shtrikh.cashbox, ll());
						const tail = kkm.egais.tail(win1251encode, receipt, ll());

						return head.reduce(printline, Promise.resolve())
							.then(qr)
							.then(() => tail.reduce(printline, Promise.resolve()));
					}

					fillmodel().
					then(function () {
						log("получаем состояние ФР");
						return talk([0x10].concat(config.operpass));
					}).
					then(function (data) {
						log("проверяем готовность ФР")
						return check(data, fail, drop, open);
					}).
					then(setcashiermaker(receipt.cashier, stalk, log, 1)).
					then(function () {
						log("открываем чек");
						return stalk([0x8d].concat(config.operpass).concat(rtype), "Ошибка открытия чека");
					}).
					then(function() {
						log("устанавливаем почту/телефон покупателя ("+receipt.counterparty+")");
						return (!!receipt.counterparty) ?
							setreq(1008, receipt.counterparty) :
							Promise.resolve();
					}).
					then(function () {
						log("печать позиций");
						return receipt.positions.reduce(sell, Promise.resolve());
					}).
					then(function () {
						log("подытог чека");
						return stalk([0x89].concat(config.operpass), "Ошибка печати подытога чека");
					}).
					then(function () {
						log("закрываем чек");
						const cash = int2bytes(receipt.paidSum, [0,0,0,0,0]).reverse();
						const nocash = int2bytes(receipt.noCashSum, [0,0,0,0,0]).reverse();
						const summ =  cash.concat(zeroarr(10)).concat(nocash.concat(zeroarr(46)));

						return stalk([0x85].concat(config.operpass).concat(summ), "Ошибка закрытия чека").
						then(function () { return new Promise(function (r) { setTimeout(r, 500); }) });
					}).
					then(function() { state.stagecritical = false; }).
					then(function () {
						log("ждем когда чек закроется");
						return wait4finish(talk);
					}).
					then(function() {
						log("печать чека ЕГАИС");
						return receipt.egais1 ? egaisprint(receipt.egais1) : Promise.resolve();
					}).
					then(function () {
						log("отрезка чека");
						return cutreceipt(stalk);
					}).
					then(function () {
						log("отключаемся");
						return new Promise(function(r) { close(r); });
					}).
					then(onfinish).
                    catch(e => {
                        close(function () { err(e.message || e); });
                    });
				}
			);
		}

		function z(cashier, onfinish) {
			app(function(send, receive, close) {
				const talk = U.talkmaker(send, receive, log);
				const stalk = U.safetalkmaker(send, receive, close, log, assert);

				setcashiermaker(cashier, stalk, log, 30)().
				then(function () { return stalk([0x41].concat(config.adminpass), "Ошибка печати Z-отчета"); }).
				then(function () { return wait4finish(talk).then(function () { return new Promise(function (r) { setTimeout(r, 200); }) }); }).
				then(function () { return new Promise(function(r) { close(r); }); }).
				then(onfinish);
			});
		}

		function x(cashier, onfinish) {
			app(function(send, receive, close) {
				const talk = U.talkmaker(send, receive, log);
				const stalk = U.safetalkmaker(send, receive, close, log, assert);

				setcashiermaker(cashier, stalk, log, 30)().
				then(function () { return stalk([0x40].concat(config.adminpass), "Ошибка печати X-отчета"); }).
				then(function () { return wait4finish(talk).then(function () { return new Promise(function (r) { setTimeout(r, 200); }) }); }).
				then(function () { return new Promise(function(r) { close(r); }); }).
				then(onfinish);
			});
		}

		function cashin(data, onfinish) {
			const summ = data.sum || data;
			const cashier = data.cashier;

			app(function(send, receive, close) {
				const talk = U.talkmaker(send, receive, log);
				const stalk = U.safetalkmaker(send, receive, close, log, assert);

				setcashiermaker(cashier, stalk, log, 1)().
				then(function () { return stalk([0x50].concat(config.operpass).concat(int2bytes(summ, [0,0,0,0,0]).reverse()), "Ошибка при внесении наличных"); }).
				then(function() { return wait4finish(talk) }).
				then(function () { return cutreceipt(stalk); }).
				then(function () { return new Promise(function(r) { close(r); }); }).
				then(onfinish);
			});
		}

		function cashout(data, onfinish) {
			const summ = data.sum || data;
			const cashier = data.cashier;

			app(function(send, receive, close) {
				const talk = U.talkmaker(send, receive, log);
				const stalk = U.safetalkmaker(send, receive, close, log, assert);

				setcashiermaker(cashier, stalk, log, 1)().
				then(function() { return stalk([0x51].concat(config.operpass).concat(int2bytes(summ, [0,0,0,0,0]).reverse()), "Ошибка при выплате наличных"); }).
				then(function() { return wait4finish(talk); }).
				then(function () { return cutreceipt(stalk); }).
				then(function () { return new Promise(function(r) { close(r); }); }).
				then(onfinish);
			});
		}

		function changext(cashier, onfinish) {
			app(
				function (send, receive, close) {
					const talk = U.talkmaker(send, receive, log);
					const stalk = U.safetalkmaker(send, receive, close, log, assert);

					setcashiermaker(cashier, stalk, log, 1)().
					then(function () { return talk([0xE0].concat(config.operpass)); }).
					then(function (data) {
					    if (data[1]) throw e0(data[1]);
					    else {
						    log("ждем когда смена откроется");
						    return wait4finish(talk);
						}
					}).then(function () {
						return new Promise(function (r) {
							log("Смена открыта");
							setTimeout(r, 200);
						});
					}).then(function () { close(onfinish); }).
					catch(function(e) {
					    close(function() { err(e.message || e, e.code); });
					});
				}
			);
		}

		function reportCalcStatusext(cashier, onfinish) {
			app(function(send, receive, close) {
				const talk = U.talkmaker(send, receive, log);
				const stalk = U.safetalkmaker(send, receive, close, log, assert);
				const settable = settablemaker(stalk);

				setcashiermaker(cashier, stalk, log, 30)().
				then(function () { return stalk([0xFF,0x38].concat(config.adminpass), "Ошибка печати отчета о состоянии рассчетов"); }).
				then(function () { return wait4finish(talk).then(function () { return new Promise(function (r) { setTimeout(r, 200); }) }); }).
				then(function () { return new Promise(function(r) { close(r); }); }).
				then(onfinish);
			});
		}

		function setreqmaker(stalk) {
			return function (req, text) {
				const textbytes = win1251encode(text);
				const reqbytes = short2bytes(req);
				const lenbytes = short2bytes(textbytes.length);

				return stalk(
					[0xff, 0x0c].
					concat(config.operpass).
					concat(reqbytes).
					concat(lenbytes).
					concat(textbytes),
					"Ошибка при установке значения ("+text+") реквизита "+req);
			}
		}

		function settablemaker(stalk) {
			return function (tab, row, field, val) {
				const rowbytes = short2bytes(row);
				const valbytes = cutbytes(win1251encode(val), 40);
				return stalk(
					[0x1e].
					concat(config.adminpass).
					concat(tab).
					concat(rowbytes).
					concat(field).
					concat(valbytes),
					"Ошибка при записи значения в таблицу ФР "+tab+", ряд "+row+", поле "+field);
			}
		}

		function setcashiermaker(cashier, stalk, log, num) {
			const settable = settablemaker(stalk);
			return (!!cashier) ?
				function () {
					log("устанавливаем имя кассира в таблица ФР для печати в чеке");
					return settable(2, num, 2, cashier);
				} :
				U.pass();
		}

		proc(beep, demo, info, type, check, printext, z, x, changext, cashin, cashout, reportCalcStatusext, corrprint, hackfr);
	});
}
kkm.shtrikh.errors = function () {
	return [
		"Ошибок нет",
		"Неисправен накопитель ФП 1, ФП 2 или часы",
		"Отсутствует ФП 1",
		"Отсутствует ФП 2",
		"Некорректные параметры в команде обращения к ФП",
		"Нет запрошенных данных",
		"ФП в режиме вывода данных",
		"Некорректные параметры в команде для данной реализации ФП",
		"Команда не поддерживается в данной реализации ФП",
		"Некорректная длина команды",
		"Формат данных не BCD",
		"Неисправна ячейка памяти ФП при записи итога",
		"",
		"",
		"",
		"",
		"",
		"Не введена лицензия",
		"Заводской номер уже введен",
		"Текущая дата меньше даты последней записи в ФП",
		"Область сменных итогов ФП переполнена",
		"Смена уже открыта",
		"Смена не открыта",
		"Номер первой смены больше номера последней смены",
		"Дата первой смены больше даты последней смены",
		"Нет данных в ФП",
		"Область перерегистраций в ФП переполнена",
		"Заводской номер не введен",
		"В заданном диапазоне есть поврежденная запись",
		"Повреждена последняя запись сменных итогов",
		"Область перерегистраций ФП переполнена",
		"Отсутствует память регистров",
		"Переполнение денежного регистра при добавлении",
		"Вычитаемая сумма больше содержимого денежного регистра",
		"Неверная дата",
		"Нет записи активизации ",
		"Область активизаций переполнена",
		"Нет активизации с запрашиваемым номером",
		"Вносимая клиентом сумма меньше суммы чека",
		"",
		"",
		"",
		"",
		"Невозможно отменить предыдущую команду",
		"Обнулённая касса (повторное гашение невозможно)",
		"Сумма чека по секции меньше суммы сторно",
		"В ФР нет денег для выплаты",
		"",
		"ФР заблокирован, ждет ввода пароля налогового инспектора",
		"",
		"Требуется выполнение общего гашения",
		"Некорректные параметры в команде",
		"Нет данных",
		"Некорректный параметр при данных настройках",
		"Некорректные параметры в команде для данной реализации ФР",
		"Команда не поддерживается в данной реализации ФР",
		"Ошибка в ПЗУ",
		"Внутренняя ошибка ПО ФР",
		"Переполнение накопления по надбавкам в смене",
		"Переполнение накопления в смене",
		"ЭКЛЗ: неверный регистрационный номер",
		"Смена не открыта – операция невозможна",
		"Переполнение накопления по секциям в смене",
		"Переполнение накопления по скидкам в смене",
		"Переполнение диапазона скидок",
		"Переполнение диапазона оплаты наличными",
		"Переполнение диапазона оплаты типом 2",
		"Переполнение диапазона оплаты типом 3",
		"Переполнение диапазона оплаты типом 4",
		"Cумма всех типов оплаты меньше итога чека",
		"Не хватает наличности в кассе",
		"Переполнение накопления по налогам в смене",
		"Переполнение итога чека",
		"Операция невозможна в открытом чеке данного типа",
		"Открыт чек – операция невозможна",
		"Буфер чека переполнен",
		"Переполнение накопления по обороту налогов в смене",
		"Вносимая безналичной оплатой сумма больше суммы чека",
		"Смена превысила 24 часа, для продолжения работы необходимо снять Z-отчет",
		"Неверный пароль",
		"Идет печать предыдущей команды",
		"Переполнение накоплений наличными в смене",
		"Переполнение накоплений по типу оплаты 2 в смене",
		"Переполнение накоплений по типу оплаты 3 в смене",
		"Переполнение накоплений по типу оплаты 4 в смене",
		"Чек закрыт – операция невозможна",
		"Нет документа для повтора",
		"ЭКЛЗ: количество закрытых смен не совпадает с ФП",
		"Ожидание команды продолжения печати",
		"Документ открыт другим оператором",
		"Скидка превышает накопления в чеке",
		"Переполнение диапазона надбавок",
		"Понижено напряжение 24В",
		"Таблица не определена",
		"Некорректная операция",
		"Отрицательный итог чека",
		"Переполнение при умножении",
		"Переполнение диапазона цены",
		"Переполнение диапазона количества",
		"Переполнение диапазона отдела",
		"ФП отсутствует",
		"Не хватает денег в секции",
		"Переполнение денег в секции",
		"Ошибка связи с ФП",
		"Не хватает денег по обороту налогов",
		"Переполнение денег по обороту налогов",
		"Ошибка питания в момент ответа по I2C ",
		"Нет чековой ленты",
		"Нет контрольной ленты",
		"Не хватает денег по налогу",
		"Переполнение денег по налогу",
		"Переполнение по выплате в смене",
		"Переполнение ФП",
		"Ошибка отрезчика",
		"Команда не поддерживается в данном подрежиме",
		"Команда не поддерживается в данном режиме",
		"Ошибка ОЗУ",
		"Ошибка питания",
		"Ошибка принтера: нет импульсов с тахогенератора",
		"Ошибка принтера: нет сигнала с датчиков",
		"Замена ПО",
		"Замена ФП",
		"Поле не редактируется",
		"Ошибка оборудования",
		"Не совпадает дата",
		"Неверный формат даты",
		"Неверное значение в поле длины",
		"Переполнение диапазона итога чека",
		"Ошибка связи с ФП",
		"Ошибка связи с ФП",
		"Ошибка связи с ФП",
		"Ошибка связи с ФП",
		"Переполнение наличности",
		"Переполнение по продажам в смене",
		"Переполнение по покупкам в смене",
		"Переполнение по возвратам продаж в смене",
		"Переполнение по возвратам покупок в смене",
		"Переполнение по внесению в смене",
		"Переполнение по надбавкам в чеке",
		"Переполнение по скидкам в чеке",
		"Отрицательный итог надбавки в чеке",
		"Отрицательный итог скидки в чеке",
		"Нулевой итог чека",
		"Касса не фискализирована",
		"Поле превышает размер, установленный в настройках",
		"Выход за границу поля печати при данных настройках шрифта",
		"Наложение полей",
		"Восстановление ОЗУ прошло успешно",
		"Исчерпан лимит операций в чеке",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"Ошибка связи с ЭКЛЗ",
		"ЭКЛЗ отсутствует",
		"ЭКЛЗ: Некорректный формат или параметр команды",
		"Некорректное состояние ЭКЛЗ",
		"Авария ЭКЛЗ",
		"Авария КС в составе ЭКЛЗ",
		"Исчерпан временной ресурс ЭКЛЗ",
		"ЭКЛЗ переполнена",
		"ЭКЛЗ: Неверные дата и время",
		"ЭКЛЗ: Нет запрошенных данных",
		"Переполнение ЭКЛЗ (отрицательный итог документа)",
		"",
		"",
		"",
		"",
		"",
		"ЭКЛЗ: Переполнение в параметре количество",
		"ЭКЛЗ: Переполнение в параметре сумма",
		"ЭКЛЗ: Уже активизирована",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"Контроль даты и времени (подтвердите дату и время)",
		"ЭКЛЗ: суточный отчёт с гашением прервать нельзя",
		"Превышение напряжения в блоке питания",
		"Несовпадение итогов чека и ЭКЛЗ",
		"Несовпадение номеров смен",
		"Буфер подкладного документа пуст",
		"Подкладной документ отсутствует",
		"Поле не редактируется в данном режиме",
		"Отсутствуют импульсы от таходатчика"
	];
};
kkm.shtrikh54fz.errors = kkm.shtrikh.errors;
if (!kkm.pirit.utils)
    kkm.pirit.utils = {};

kkm.pirit.utils.esc = function (data) {
    var i = -1;
    while (++i < data.length)
        if (ETX == data[i] || DLE == data[i])
            data.splice(i++, 0, DLE);
}

kkm.pirit.utils.unesc = function (data) {
    var i = 0;
    while(i++<data.length)
        if (data[i] == ACK)
            data.splice(i,1);
    return data;
}

kkm.pirit.utils.piritcrc = function (crc) {
    i1 = Math.floor(crc / 16);
    if(i1 > 9) i1 = i1 + 0x37;
    else i1 = i1 | 0x30;
    i2 = (crc % 16);
    if(i2 > 9) i2 = i2 + 0x37;
    else i2 = i2 | 0x30;

    return [i1, i2];
}

kkm.pirit.utils.pack = function (data) {
    const PU = kkm.pirit.utils;
    const crc = kkm.util.crc;
    const piritcrc = PU.piritcrc;

    var i = 0;
    var i1, i2;
    i1 = 0; i2 = 0;

    PU.esc(data);
    data.push(ETX);
    i = piritcrc(crc(data));
    data.push(i[0]);
    data.push(i[1]);
    data.unshift(STX);
    log("Data to FR:"+data);
    return data;
}

kkm.pirit.utils.unpack = function (d) {
    const PU = kkm.pirit.utils;

    const shft = firstIndex(d, d.length, function(i, v) {
      if (STX != v) return false;
      if (d.length -i < 8) return false;
      if (PU.sentid != d[i+1]) return false;

      return true;
    });
    const limit = shft + firstIndex(d.slice(shft), d.length-shft, function(i, v) { return ETX == v; });

    const data = d.slice(shft, limit + 3);

    if (STX != data[0])
        throw new Error ("no STX found");
    data.shift();
    if (ETX != data[data.length - 3])
        throw new Error("no ETX found");
    data.pop();
    data.pop();
    data.pop();
    return PU.unesc(data);
}

kkm.pirit.utils.receivedpacket = function (b, l) {
    const PU = kkm.pirit.utils;

    const crc = kkm.util.crc;
    const piritcrc = PU.piritcrc;

    if (ENQ == b[l-1]) return true;

    const shft = firstIndex(b, l, function(i, v) {
      if (STX != v) return false;
      if (l -i < 8) return false;
      if (PU.sentid != b[i+1]) return false;

      return true;
    });

    if (-1 == shft) return false;

    const lmt = firstIndex(b, l, function(i, v ){
        if (i < shft+5) return false;
        if (v != ETX) return false;
        if ((i + 2) >= l) return false;

        const сcrc = piritcrc(crc(b.slice(shft+1, i+1)));
        return (сcrc[0] == b[i+1] && сcrc[1] == b[i+2]);
    });

    return (-1 < lmt);
}

kkm.pirit.utils.sentid = false;

kkm.pirit.command = function (config, client, logging, conn, send1, receive1, close, flush) {
    kkm.util.incl(kkm.util.modlog(logging(), '......'));
    const bytechecker = kkm.util.bytecheck;
    const crc = kkm.util.crc;
    const PU = kkm.pirit.utils;

    PU.sentid = false;

    function r(conn, check, callback) {
        receive1(conn, check, function (d) {
            flush(conn, function (){
                callback(d);
            });
        });
    }

    var lastsend = function() { log("nothing been sent yet to pirit"); };
    function send (d, sentlink) {
      const data = PU.pack(d);
        (lastsend = function() {
            log("Function Send Start");
                send1(conn, data, function(si) {
                  PU.sentid = data[5];
                  const le = chrome.runtime.lastError;
                  var msg = "";
                  if((!si || si.error))
                    if(le && le.message) msg = le.message;
                    else msg = "data sent failed!";
                  else msg = "data sent!";
                  log(msg);
                  if (sentlink)
                    sentlink(send, receive);
                });
                log("Function Send End");
        })();
    };

    function receive(receivedlink) {
        log("Function Recieve Start");
        var timeout = false;
        function retry(attempt) {
            timeout = setTimeout(function() {
                if (config.retries < attempt) log("command response receive failed in "+config.retries+" attemts");
                else {
                    log("resending last sent command, attempt "+attempt+" of "+config.retries)
                    lastsend();
                    retry(attempt + 1);
                }
            }, 2000);
        }

        r(conn, PU.receivedpacket, function(d) {
            clearTimeout(timeout);
            if (ENQ == d[d.length - 1]) send1(conn, [ACK], function() { receive(receivedlink); });
            else receivedlink(PU.unpack(d));
            log("Function Recieve END");
        });
        log("receive initted");

        retry(1);
    }

    client(send, receive, close);
}
if (!kkm.pirit.utils)
    kkm.pirit.utils = {};

kkm.pirit.utils.errcode = function (data) {
    if (!data || !data.length || data.length < 5) return 0;
    return 10 * (data[3]-0x30) + (data[4]-0x30);
}

kkm.pirit.utils.statusmaker = function (log) {
    return function (data) {
        var mas = new Array();
        var o1 = 0;
        var f1 = 0;
        var f2 = 0;
        var f3 = 0;
        log("Fuction getPiritStatus Start");
        o1 = data.indexOf(28);
        log("O = "+o1);

        mas = data.slice(5,o1);
        for(i=0;i<mas.length; i++)
        {
            f1 *= 10;
            f1 += mas[i] - 0x30;
        }

        o2 = data.indexOf(28,o1+1);
        mas = data.slice(o1+1,o2);
        for(i=0;i<mas.length; i++)
        {
            f2 *= 10;
            f2 += mas[i] - 0x30;
        }
        for(i=o2+1;i<data.length-1; i++)
        {
            f3 *= 10;
            f3 += data[i] - 0x30;
        }
        log("f1 = "+f1);
        log("f2 = "+f2);
        log("f3 = "+f3);

        mas[0] = f1;
        mas[1] = f2;
        mas[2] = f3;
        log("Fuction getPiritStatus End");
        return mas;
    }
};
kkm.pirit.utils.link = function (send, receive, config) {
    function p(cmd) {
        return ((config.password).concat(cmd));
    }

    return function (data, save) {
        return function () {
            return new Promise(function (next) {
                receive(function (d) {
                    save(d);
                    next(d);
                });
                send(p(data), function () {});
            });
        };
    }
};
kkm.pirit.utils.commandStart = function (send, receive, close, config, func1) {
    var  CurrTime = new Date();
    ctday = CurrTime.getDate();
    ctday = (ctday>9)?ctday:"0"+ctday;

    ctmon = CurrTime.getMonth()+1;
    ctmon = (ctmon>9)?ctmon:"0"+ctmon;

    ctyear = CurrTime.getYear();
    ctyear = (ctyear%100);

    cthour = CurrTime.getHours();
    cthour = (cthour>9)?cthour:"0"+cthour;

    ctmin = CurrTime.getMinutes();
    ctmin = (ctmin>9)?ctmin:"0"+ctmin;

    ctsec = CurrTime.getSeconds();
    ctsec = (ctsec>9)?ctsec:"0"+ctsec;

    str = ""+ctday+ctmon+ctyear+String.fromCharCode(0x1C)+cthour+ctmin+ctsec;

    const PU = kkm.pirit.utils;
    const l = PU.link(send, receive, config);

    l([0x24, 0x31, 0x30].concat(charencode(str)), func1)();
};
kkm.pirit.utils.frerrormaker = function(log, err) {
    return function(f1) {
        if (f1 == 0) return;
        switch(true) {
            case ((f1 &   1) != 0): err("Неверная контрольная сумма NVR");break;
            case ((f1 &   2) != 0): err("Неверная контрольная сумма в конфигурации");break;
            case ((f1 &   4) != 0): err("Ошибка интерфейса с ФП");break;
            case ((f1 &   8) != 0): err("Неверная контрольная сумма фискальной памяти");break;
            case ((f1 &  16) != 0): err("Ошибка записи в фискальную память");break;
            case ((f1 &  32) != 0): err("Фискальный модуль не авторизован");break;
            case ((f1 &  64) != 0): err("Фатальная ошибка ЭКЛЗ");break;
            case ((f1 & 128) != 0): err("Расхождение данных ФП и ЭКЛЗ");break;
        }
        log("Error from FR:"+f1);
    }
};
kkm.pirit.utils.convIntToInt = function (intNum) {
    const res = [0];
    str = intNum.toString();
    len = str.length;
    for (var i=0; i<len; i++)
    {
        if((str[i] == ".")||(str[i]==",")) res[i]=0x2E;
        else if((str[i] >= "0")&&(str[i]<="9")) res[i]= 0x30+parseInt(str[i],10);;
    }
    return res;
}

kkm.pirit.utils.version = 0;

kkm.pirit.api = function (logging, config, executor, proc) {
    const U = kkm.util;

    const empty = U.empty;
    const map = U.map;
    const crc = U.crc;

    const L = U.modlog(logging(), "...");

    const log = L.log;
    const err = L.err;

    const assert = U.assertmaker(config.errtable, err);

    const PU = kkm.pirit.utils;
    const getPiritStatus = PU.statusmaker(log);
    const frerror = PU.frerrormaker(log, err);

    function app(chain) {
        executor(logging, chain);
    }

    function p(cmd) {
        return ((config.password).concat(cmd));
    }

    const talkmaker = function(send, receive, log) {
        const talk = talkmaker(send, receive, log);
        return function(data) {
            talk(p(data));
        }
    };

    const safetalkmaker = function(send, receive, close, log, assert) {
        const stalk = safetalkmaker(send, receive, close, log, assert);
        return function(data, assertmsg) {
            return stalk(p(data), assertmsg);
        }
    };


    function beep(data, onfinish) {
        const PU = kkm.pirit.utils;
        function doBeep(send, receive, close)
        {

            send(p([0x81, 0x30, 0x30]), function() {
                log("beeeeep.............................................Start");
                receive(function (data){
                    log("Hello World!!!!!!!!!!!!!!!!!!!!!!!!!!");
                    for(i=0;i<data.length; i++)
                    {
                        log("D["+i+"]:"+data[i]);
                    }
                    var m10 = new Array();
                    m10 = getPiritStatus(data);

                    log("1="+m10[0]+"  2="+m10[1]+ "   3="+m10[2]);
                    f1 = m10[0];
                    f2 = m10[1];
                    f3 = m10[2];

                    var errNum = PU.errcode(data);

                    if (errNum != 0) log("Error from FR B1:"+errNum);
                    if (f1 != 0) log("Error from FR:"+f1);
                    if ((f2 & 1) != 0){
                        PU.commandStart(send, receive, close, config, function (){doBeep(send, receive, close);});
                    }else{
                        send(p([0x82, 0x38, 0x32, 0x37, 0x38]), function() {
                            receive(function (data){
                                close();
                            });
                        });
                    }
                    if (f3 != 0) log("Error from FR B3:"+f3);

                });
                log("beeeeep..............................................End");
            });

        }
        app(
            function (send, receive, close) {
                doBeep(send, receive, close);
            }
        );
    }

    function demo(data, onfinish) {
        app(function (send, receive, close) {
            const PU = kkm.pirit.utils;
            const l = PU.link(send, receive, config);

            l([0x94, 0x39, 0x34], function () {
                close();
                if (onfinish && onfinish instanceof Function)
                    onfinish();
            })();
        });
    }

    function demotext(data, onfinish) {
      log("demo print");
      app(function(send, receive, close) {

        function demoprint() {
          const text = cutbytes(cp866encode(kkm.config.default.demotext), 72);
          const opendoc= cp866encode("30")
          .concat(cp866encode((0b00000001).toString(16)))
          .concat([0x1C, 0x31, 0x1C, 0x8A, 0xA0, 0xE1, 0xE1, 0xA8, 0xE0, 0x1C, 0x30, 0x1C]);
          const printtextcom = cp866encode("40");
          const closedoc = cp866encode("31");
          return talk(p([0x50].concat(opendoc))).
          then(function() { return talk(p([0x51].concat(printtextcom).concat(text).concat([0x1C]).concat(cp866encode((0b00000001).toString(16))))); }).
          then(function() { return talk(p([0x51].concat(printtextcom).concat(cutbytes(cp866encode(""), 72)).concat([0x1C]).concat(cp866encode((0b00000001).toString(16))))); }).
          then(function() { return talk(p([0x52].concat(closedoc))); }).
          then(
            function ()
                    {
                        close();
                        if (onfinish && onfinish instanceof Function)
                            onfinish();
                    }
          );
        }

        function check(data) {
            const PU = kkm.pirit.utils;
            const retry = checkFR;
            const m10 = getPiritStatus(data);
            frerror(m10[0]);
            if ((m10[1] & 1) != 0) PU.commandStart(send, receive, close, config, retry);
            else if (m10[2] != 0) talk(p([0x32, 0x33, 0x32])).then(retry);
            else return demoprint();
        }
        const talk = U.talkmaker(send, receive, log);

        function checkFR()
        {
          log("проверяем состояние ФР");
                talk(p([0x81, 0x30, 0x30])).
                then(check);
        }

        checkFR();

      });
    }

    function info(proc) {
        app(function (send, receive, close) {
            inforeq(send, receive, close, function(data) {
                close(function () { proc(data); });
            });
        });
    }

    function inforeq(send, receive, proc) {
        const res = {};
        function s(path) {
            return function (data) {
                res[path] = data;
            }
        }

        const PU = kkm.pirit.utils;
        const l = PU.link(send, receive, config);

        l([0x25, 0x30, 0x32, 0x34], s("num"))().
        then(l([0x27, 0x30, 0x30], s("state"))).
        then(l([0x28, 0x30, 0x31, 0x31], s("shift"))).
        then(l([0x29, 0x30, 0x31, 0x32], s("receipt"))).
        catch(function(e) { err(e); }).
        then(function () {
            proc(res);
        });
    }

    function type(proc) {
        app(function (send, receive, close) {
            typereq(send, receive, function(data) {
                close(function() { proc(data); });
            });;
        });
    }

    function typereq(send, receive, proc) {
        const PU = kkm.pirit.utils;
        const l = PU.link(send, receive, config);
        l([0x26, 0x30, 0x32, 0x32], function (data) { proc(data); })().
        then(function () { log("type info got"); }).
        catch(function (e) { err("error requesting type info: "+e); });
    }

    function model54fz(fw) {
        switch (true) {
            case (fw >=  10 && fw <  200): return ["Пирит ФР-01К ЕНВД", false];
            case (fw >= 200 && fw <= 250): return ["Pirit-K", false];
            case (fw >= 400 && fw <= 450): return ["Pirit ЕНВД", false];
            case (fw >= 500 && fw <= 549): return ["Pirit 01Ф", true];
            case (fw >= 550 && fw <= 599): return ["Pirit 02Ф", true];
            case (fw >= 600 && fw <= 650): return ["VikiPrint", false];
            case (fw >= 650 && fw <= 699): return ["VikiPrint Ф", true];
            default:
                log("Unknown firmware: " + fw + ". Setting default Pirit")
                return ["Pirit", false];
        }
    }

    function check(res, checked) {
        const datas = [];
        app(
            function (send, receive, close) {
                typereq(send, receive, function (data) {
                    log(data[3] === 48 && data[4] === 48 ? "type data received" : "error " + win1251decode([data[3], data[4]]) );
                    datas.push(data);

                    const firmware = win1251decode(data.slice(data.indexOf(28, 5) + 1, data.indexOf(28, data.indexOf(28, 5) + 1)));
                    log("firmware :"+firmware);
                    const mAndFz = model54fz(firmware);
                    res.ver = " v. "+firmware;
                    kkm.pirit.utils.version = firmware;
                    res.model = mAndFz[0];
                    res.fz54 = mAndFz[1];
                    res.vendor = res.fz54 ? "pirit54fz" : "pirit";

                    inforeq(send, receive, function(data) {
                        log("info data received! ["+data+"]");
                        datas.push(data);

                        res.fiscal = Boolean(!(data.state.slice(3, data.state.indexOf(28))[0] & 1));
                        res.doc = 0;
                        res.receipt = chardecode(data.receipt.slice(3, data.receipt.indexOf(28)));
                        res.change = chardecode(data.shift.slice(3, data.shift.indexOf(28)));
                        res.serial = chardecode(data.num.slice(7,18));

                        close(function () { checked(res, datas); });
                    });
                });
            }
        );
    }


    function convIntToInt(intNum)
    {
        const res = [0];
        str = intNum.toString();
        len = str.length;
        for (var i=0; i<len; i++)
        {
            if((str[i] == ".")||(str[i]==",")) res[i]=0x2E;
            else if((str[i] >= "0")&&(str[i]<="9")) res[i]= 0x30+parseInt(str[i],10);;
        }
        return res;
    }

    function vat(v) {
        switch(v) {
            case (18): return 0x30;
            case (10): return 0x31;
            case (0): return 0x32;
            default: return 0x33;
        }
    }

    function print(receipt, onfinish) {
        function rtype() {
            return receipt.return ? 0x33: 0x32;
        }


        function sellt(send, receive, close, rl, link, maxlength)
        {
            log("Function Sellt");
            log("печать строки "+JSON.stringify(rl));

            var length = rl.name.length > maxlength ? maxlength : rl.name.length;

            const name = splitbytes(cp866encode(rl.name.substring(0, length)), 224);
            const nameEx = cp866encode_pirit(rl.name.substring(0,length));

            strName = nameEx;
            str1C = [0x1C];
            strQ = convIntToInt(rl.quantity);
            strP = convIntToInt(rl.price/100);
            strVat = vat(rl.vat);


            strM = [0x42, 0x34, 0x32].concat(strName).concat(str1C).concat(str1C).concat(strQ).concat(str1C).concat(strP).concat(str1C).concat(strVat).concat(str1C).concat(str1C).concat(0x31).concat(str1C);

            map(name, function (v, f) {
                send(p(strM), function() {
                    receive(function (data){
                        log("позиция добавлена");
                        f();
                    });
                });
            }, link);
        }

        const emsg = function(code) {
            return config.errtable[code] || "Ошибка при открытии смены ("+code+")";
        };

        function printReceipt(send, receive, close, receipt, onfinish) {
            const PU = kkm.pirit.utils;
            const l = PU.link(send, receive, config);
            function e(m) {
                throw new Error(m);
            }
            var toprint;

            return new Promise(function (f) { log("starting print receipt..."); f(); }).
            then(l([0x81, 0x30, 0x30], function (data) {
                var m10 = new Array();
                m10 = getPiritStatus(data);

                f1 = m10[0];
                f2 = m10[1];
                f3 = m10[2];

                var errNum = PU.errcode(data)

                if (errNum != 0) log("Error from FR err:"+errNum);
                if (f1 != 0) {
                    log("Error from FR B1:"+f1);
                    if ((f1 &   1) != 0) e("Неверная контрольная сумма NVR");
                    if ((f1 &   2) != 0) e("Неверная контрольная сумма в конфигурации");
                    if ((f1 &   4) != 0) e("Ошибка интерфейса с ФП");
                    if ((f1 &   8) != 0) e("Неверная контрольная сумма фискальной памяти");
                    if ((f1 &  16) != 0) e("Ошибка записи в фискальную память");
                    if ((f1 &  32) != 0) e("Фискальный модуль не авторизован");
                    if ((f1 &  64) != 0) e("Фатальная ошибка ЭКЛЗ");
                    if ((f1 & 128) != 0) e("Расхождение данных ФП и ЭКЛЗ");
                } else if ((f2 & 1) != 0) {
                    toprint = function (onfinish) { PU.commandStart(send, receive, close, config, function () { printReceipt(send, receive, close, receipt, onfinish); }); };
                } else if ((f2 & 8) != 0) {
                    err("Смена больше 24 часов");
                    close();
                } else if (f3 != 0){
                    toprint = function (onfinish) {
                        receive(function (data){
                            printReceipt(send, receive, close, receipt, onfinish);
                        });
                        send(p([0x32, 0x33, 0x32]), function() {});
                    };
                } else {
                    toprint = function (onfinish) {
                        receive(function (data) {
                            const ec = PU.errcode(data)
                            if (ec) return close(()=>err(emsg(ec, 001)));

                            masPos = receipt.positions;
                            map(receipt.positions, function(v, f) {
                                l([0x26, 0x30, 0x32, 0x32], function (data) {
                                    var firmware = win1251decode(data.slice(data.indexOf(28, 5) + 1, data.indexOf(28, data.indexOf(28, 5) + 1)));
                                    var maxlength = 224;
                                    if (firmware >= 10 && firmware < 200) {
                                        maxlength = 150;
                                    }
                                    sellt(send, receive, close, v, f, maxlength);
                                })()

                            }, function() {
                                log("печать строк завершена");
                                receive(function (){
                                    const t = U.talkmaker(send, receive, log);
                                    const exec = (strM) => t(p(strM));

                                    function printSum(summ, type) {
                                        log("печатаем сумму");

                                        const strD = convIntToInt(summ/100);
                                        const strM = [0x47, 0x34, 0x37].concat(type).concat(0x1C).concat(strD).concat(0x1C).concat(0x1C);

                                        return exec(strM);
                                    }

                                    function closeReceipt() {
                                        log("закрытие чека");
                                        const enccounterparty = !!receipt.counterparty ? cp866encode(receipt.counterparty).concat([0x1C]) : [];

                                        const strM = [0x31, 0x33, 0x31, 0x30, 0x1C].concat(enccounterparty);

                                        return exec(strM);
                                    }

                                    function egaisprint() {
                                        function opendoc() {
                                            log("Открываем сервисный документ");

                                            const strM = [0x28]
                                                .concat(cp866encode("30"))
                                                .concat(cp866encode((0b00000001).toString(16)))
                                                .concat([0x1C, 0x31, 0x1C, 0x8A, 0xA0, 0xE1, 0xE1, 0xA8, 0xE0, 0x1C, 0x30, 0x1C]);

                                            return exec(strM);
                                        }

                                        function readcashbox() {
                                            log("Читаем номер кассы");

                                            const strM = [0x27]
                                                .concat(cp866encode("11"))
                                                .concat(cp866encode("10"))
                                                .concat([0x1c])
                                                .concat(cp866encode("0"))
                                                .concat([0x1c]);

                                            return exec(strM).then((data) => {
                                                let cashbox = cp866decode(data.slice(5, data.indexOf(0x1c)));
                                                log("Номер кассы: " + cashbox);
                                            });
                                        }

                                        function print(text) {
                                            log("Печать строки: " + JSON.stringify(text));

                                            const strM = [0x29]
                                                .concat(cp866encode("40"))
                                                .concat(text)
                                                .concat([0x1c])
                                                .concat(cp866encode((0b00000001).toString(16)));

                                            return exec(strM);
                                        }

                                        function printarr(arr) {
                                            return arr.reduce((prev, e, index) => {
                                                return prev.then(() => print(e));
                                            }, Promise.resolve());
                                        }

                                        function qr() {
                                            log("Печать QR-кода");

                                            const zero = cp866encode("0");
                                            const strM = [0x30]
                                                .concat(cp866encode("41"))
                                                .concat(zero)
                                                .concat([0x1c])
                                                .concat(cp866encode("6"))
                                                .concat([0x1c])
                                                .concat(zero)
                                                .concat([0x1c])
                                                .concat(cp866encode("8"))
                                                .concat([0x1c])
                                                .concat(cp866encode(receipt.egais1.url))
                                                .concat([0x1c]);

                                            return exec(strM);
                                        }

                                        function closedoc() {
                                            log("Закрываем документ");

                                            const strM = [0x31]
                                                .concat(cp866encode("31"));

                                            return exec(strM)
                                        }

                                        function doPrint() {
                                            log("печать чека УТМ");

                                            const ll = 40;
                                            const tail = kkm.egais.tail(cp866encode, receipt.egais1, ll);

                                            return opendoc().
                                                then(() => readcashbox()).
                                                then((cashbox) => kkm.egais.head(cp866encode, receipt.egais1, 1, ll)).
                                                then((head) => printarr(head)).
                                                then(() => qr()).
                                                then(() => printarr(tail)).
                                                then(closedoc);
                                        }

                                        return doPrint();
                                    }

                                    function regdiscount() {
                                        log("скидка");

                                        const strD = convIntToInt(receipt.discount / 100);
                                        const strM = [0x45, 0x34, 0x35, 0x31, 0x1C, 0x1C].concat(strD);

                                        return exec(strM);
                                    }


                                    (typeof receipt.discount != "undefined" ? regdiscount() : Promise.resolve()).
                                    then(receipt.noCashSum ? () => printSum(receipt.noCashSum, 0x31) : Promise.resolve()).
                                    then(receipt.paidSum ? () => printSum(receipt.paidSum, 0x30) : Promise.resolve()).
                                    then((!receipt.paidSum && !receipt.noCashSum)? () => printSum(0, 0x30) : Promise.resolve()).
                                    then(() => closeReceipt()).
                                    then(receipt.egais1 ? () => egaisprint() : Promise.resolve()).
                                    then(() => {
                                            close();
                                            if (onfinish)
                                                onfinish();
                                        });
                                });
                                send(p([0x44, 0x34, 0x34]), function() {
                                    log("предварительный итог");
                                });
                            });
                        });
                        const enccashier = !!receipt.cashier ? cp866encode(receipt.cashier) : [0x8A, 0xA0, 0xE1, 0xE1, 0xA8, 0xE0];
                        send(p([0x30, 0x33, 0x30, rtype(), 0x1C, 0x31, 0x1C].concat(enccashier).concat([0x1C, 0x30, 0x1C])), function() {});
                    };
                }
            })).
            then(function () { return new Promise(function (next) { if (toprint instanceof Function) toprint(next); else next(); }); }).
            then(function () { onfinish(); log("...print receipt finished"); }).
            catch(function (e) { close (function () { err(e.message || e); }); });
        }

        app(function (send, receive, close) {
            printReceipt(send, receive, close, receipt, onfinish);
        });
    }

    function printReport(send, receive, close, onfinish, repcode, cashier) {
        const PU = kkm.pirit.utils;

        receive(function (data) {
            var m10 = new Array();
            m10 = getPiritStatus(data);

            log("1="+m10[0]+"  2="+m10[1]+ "   3="+m10[2]);
            f1 = m10[0];
            f2 = m10[1];
            f3 = m10[2];

            var errNum = PU.errcode(data);

            if (errNum != 0) log("Error from FR err:"+errNum);
            if (f1 != 0) {
                frerror(f1);
            } else if ((f2 & 1) != 0) {
                PU.commandStart(send,receive, close, config, function(){printReport(send, receive, close, onfinish, repcode, cashier);});
            } else if ((f2 & 4) == 0){
                err("Смена не открыта");
                close();
            } else if (f3 != 0){
                receive(function (data){
                    printReport(send, receive, close, onfinish, repcode, cashier);
                });
                send(p([0x32, 0x33, 0x32]), function() {
                    log("Error from FR B3:"+f3);
                });
            } else {
                receive(function (data) {
                    close();
                    if (onfinish)
                        onfinish();
                });
                const enccashier = !!cashier ? cp866encode(cashier) : [0x8A, 0xA0, 0xE1, 0xE1, 0xA8, 0xE0];
                send(p([0x21, 0x32, repcode].concat(enccashier)), function() {});
            }
            log("report...........................................END");
        });
        send(p([0x81, 0x30, 0x30]), function() {
            log("report.............................................Start");
        });

    }

    function z(data, onfinish) {
        app(function(send, receive, close) {
            printReport(send, receive, close, onfinish, 0x031, data);
        });
    }

    function x(data, onfinish) {
        app(function(send, receive, close) {
            printReport(send, receive, close, onfinish, 0x030, data);
        });
    }

    function change(data, onfinish) {
        if (onfinish && onfinish instanceof Function) onfinish();
    }

    function cashopmaker(opcode, data) {
        return function (datas, onfinish) {
            function printCashop(send, receive, close, onfinish) {
                const PU = kkm.pirit.utils;
                const l = PU.link(send, receive, config);

                log("проверка состояния ФР перед "+ (opcode == 0x34 ? "внесением" : "выплатой"));
                receive(function (data) {
                    log("получили состояние");
                    var m10 = new Array();
                    m10 = getPiritStatus(data);

                    f1 = m10[0];
                    f2 = m10[1];
                    f3 = m10[2];

                    var errNum = PU.errcode(data);

                    if (errNum != 0) log("Error from FR err:"+errNum);
                    if (f1 != 0) {
                        log("Error from FR B1:"+f1);
                        if ((f1 &   1) != 0) e("Неверная контрольная сумма NVR");
                        if ((f1 &   2) != 0) e("Неверная контрольная сумма в конфигурации");
                        if ((f1 &   4) != 0) e("Ошибка интерфейса с ФП");
                        if ((f1 &   8) != 0) e("Неверная контрольная сумма фискальной памяти");
                        if ((f1 &  16) != 0) e("Ошибка записи в фискальную память");
                        if ((f1 &  32) != 0) e("Фискальный модуль не авторизован");
                        if ((f1 &  64) != 0) e("Фатальная ошибка ЭКЛЗ");
                        if ((f1 & 128) != 0) e("Расхождение данных ФП и ЭКЛЗ");
                    } else if ((f2 & 8) != 0) {
                        err("Смена больше 24 часов");
                        close();
                    }
                    else if (f3 != 0){
                       err("В ККТ есть открытый документ");
                       close();
                    }
                    else {
                        log("открываем документ "+ (opcode == 0x34 ? "внесения" : "выплаты"));
                        receive(function (data) {
                            log("отправляем сумму "+ (opcode == 0x34 ? "внесения" : "выплаты"));
                            receive(function (data) {
                                log("закрываем документ "+ (opcode == 0x34 ? "внесения" : "выплаты"));
                                receive(function (data) {
                                    log(0x34 == opcode ? "внесение завершено": "выплата завершена");
                                    close(onfinish);
                                });
                                send(p([0x31, 0x33, 0x31, 0x30, 0x1C]));
                            });
                            const sum = !!datas.sum ? datas.sum : datas;
                            send(p([0x48, 0x34, 0x38, 0x1C].concat(convIntToInt(sum/100.0))));
                        });
                        const enccashier = !!datas.cashier ? cp866encode(datas.cashier) : [0x8A, 0xA0, 0xE1, 0xE1, 0xA8, 0xE0];
                        send(p([0x30, 0x33, 0x30, opcode, 0x1C, 0x31, 0x1C].concat(enccashier).concat([0x1C, 0x30, 0x1C])));
                    }
                });
                send(p([0x81, 0x30, 0x30]));
            }

            app(function (send, receive, close) {
                printCashop(send, receive, close, onfinish);
            });
        }
    }

    function cashin(data, onfinish)
    {
      cashopmaker(0x34, data)(data, onfinish);
    }

    function cashout(data, onfinish)
    {
      cashopmaker(0x35,data)(data, onfinish);
    }

    function ofdreport(data, onfinish) {
        onfinish();
    }

    proc(beep, demotext, info, type, check, print, z, x, change, cashin, cashout, ofdreport);
 }
kkm.pirit.ext = function (logging, config, executor, proc) {
	kkm.pirit.api(logging, config, executor, function (beep, demo, info, type, check, print, z, x, change, cashin, cashout) {
	    const U = kkm.util;
	    const L = U.modlog(logging(), "...");

	    const log = L.log;
	    const err = L.err;

	    const X = kkm.ext;

		proc(beep, demo, info, type, check, print, z, x, change, cashin, cashout, U.cbpass, U.cbpass, U.cbpass);
	});
}
kkm.pirit.errors = function () {
	return [
		"Ошибок нет",
		"Функция невыполнима при данном статусе ККТ",
		"В команде указан неверный номер функции",
		"Некорректный формат или параметр команды",
		"Переполнение буфера коммуникационного порта",
		"Таймаут при передаче байта информации",
		"В протоколе указан неверный пароль",
		"Ошибка контрольной суммы в команде",
		"Конец бумаги",
		"Принтер не готов",
		"Текущая смена больше 24 часов",
		"Разница во времени, ККТ и указанной в команде начала работы, больше 8 минут",
		"Вводимая дата более ранняя, чем дата последней фискальной операции",
		"Неверный пароль доступа к ФП",
		"Отрицательный результат",
		"Для выполнения команды необходимо закрыть смену",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"Фатальная ошибка ККТ",
		"Нет свободного места в фискальной памяти ККТ",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"Некорректный  формат или параметр команды ЭКЛЗ",
		"Некорректное состояние ЭКЛЗ",
		"Авария ЭКЛЗ",
		"Авария КС (Криптографического сопроцессора) в составе ЭКЛЗ",
		"Исчерпан временной ресурс использования ЭКЛЗ",
		"ЭКЛЗ переполнена",
		"Неверные дата или время",
		"Нет запрошенных данных",
		"Переполнение (отрицательный итог документа, слишком много отделов для клиента)",
		"Нет ответа от ЭКЛЗ",
		"Ошибка при обмене данными с ЭКЛЗ",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"",
		"tttt"
	];
};
kkm.pirit54fz.command = kkm.pirit.command;
kkm.pirit54fz.ext = function (logging, config, executor, proc) {
	kkm.pirit.ext(logging, config, executor, function (beep, demo, info, type, check, print, z, x, change, cashin, cashout, ofdreport, corrprint, hackfr) {
	    const U = kkm.util;
	    const L = U.modlog(logging(), "...");

	    const log = L.log;
	    const err = L.err;

	    const X = kkm.ext;
	    const PU = kkm.pirit.utils;

	    const status = PU.statusmaker(log);
	    const frerror = PU.frerrormaker(log, err);

	    const e0 = function(code) {
	        const msg = config.errtable[code] || "Ошибка при открытии смены ("+code+")";
	        const e = new Error(msg);
	        e.code = 001;
	        return e;
	    };

	    function app(chain) {
	        executor(logging, chain);
	    }

	    const talkmaker = function(send, receive, log) {
	    	const t = U.talkmaker(send, receive, log);
	    	return function(data) {
	    		return t(config.password.concat(data));
	    	}
	    }

	    const talkfunctionmaker = function(log, talk) {
	    	return function(msg, data) {
	    		return function () {
		    		log(msg);
		    		return talk(data);
	    		}
	    	}
	    }

	    function getcashier(data) {
	    	return (!!data ? cp866encode(data) : [0x8A, 0xA0, 0xE1, 0xE1, 0xA8, 0xE0]).concat(0x1c);
	    }

	    function changeext(data, onfinish) {
	    	const PU = kkm.pirit.utils;
	    	app(function(send, receive, close){

				function changeprint() {
					return talk([0x29, 0x32, 0x33].concat(getcashier(!!data && !!data.cashier ? data.cashier : false))).
					then(function (data) {
						const ec = PU.errcode(data);
					    if (ec) throw e0(ec);
					    else close(onfinish);
					}).
	    			catch(function(e) { close(err(e.message || e, e.code)); });
				}

		    	function check(data) {
		    		const retry = checkFR;

		    		const m10 = status(data);
		    		frerror(m10[0]);
		    		if ((m10[1] & 1) != 0) PU.commandStart(send, receive, close, config, retry);
		    		else if ((m10[1] & 4) != 0) close(err("Смена уже открыта"));
		    		else if (m10[2] != 0) talk([0x32, 0x33, 0x32]).then(retry);
		    		else return changeprint();
		    	}

	    		const talk = talkmaker(send, receive, log);
	    		const talkf = talkfunctionmaker(log, talk);

				function checkFR() {
		          log("проверяем состояние ФР");
		        		talk([0x81, 0x30, 0x30]).
		        		then(check);
		        }
		        checkFR();
	    	});
	    }

	    function ofdreportext(data, onfinish) {
	    	const PU = kkm.pirit.utils;
	    	app(function(send, receive, close) {

				function ofdreportprint() {
					return talk([0x30, 0x35, 0x39].concat(cashier)).
					then(function() { close(onfinish); }).
					catch(function(e) { close(err(e.message || e)); });
				}

	    		const retry = checkFR;

		    	function check(data) {
		    		const m10 = status(data);
		    		frerror(m10[0]);

		    		if ((m10[1] & 1) != 0) PU.commandStart(send, receive, close, config, retry);
		    		else if ((m10[1] & 4) != 0) close(err("Смена открыта"));
		    		else return ofdreportprint();
		    	}

	    		const talk = talkmaker(send, receive, log);
	    		const talkf = talkfunctionmaker(log, talk);

	    		const cashier = getcashier(data);

				function checkFR() {
			      log("проверяем состояние ФР");
			    		talk([0x81, 0x30, 0x30]).
			    		then(check);
			    }

			    checkFR();
	    	});
	    }

		function corrprint(prms, onfinish) {
			const PU = kkm.pirit.utils;
			if(!PU.version || PU.version < 158 || (PU.version>=550 && PU.version<559) || (PU.version>=650 && PU.version<659))
			{
				return err("Печать чека коррекции для прошивки "+PU.version+" невозможна. Обновите прошивку на последнюю актуальную версию.");
			}
			app(function(send, receive, close) {
				const talk = talkmaker(send, receive, log);
				const commandStartPromise = function() {
					return new Promise(function(res) {
						PU.commandStart(send, receive, close, config, res);
					});
				}

				talk([0x81, 0x30, 0x30]).
				then(function(data) {
					const m10 = status(data);
					frerror(m10[0]);
					return Promise.resolve(m10);
				}).then(function(m10) {
					if ((m10[1] & 1) == 0) return Promise.resolve(m10);
					return commandStartPromise().
					then(function() { return Promise.resolve(m10); });
				}).then(function (m10) {
					if ((m10[1] & 4) != 0) {
						if ((m10[1] & 8) != 0) throw "Смена больше 24 часов";
						return Promise.resolve(m10);
					}
					return talk([0x29, 0x32, 0x33].concat(getcashier(!!prms && !!prms.cashier ? prms.cashier : false))).
					then(function (data) {
						  const ec = PU.errcode(data);
					    if (ec) throw e0(ec);
						else return Promise.resolve(m10);
					});
				}).then(function(m10) {
					if (m10[2] == 0) return Promise.resolve();
					return talk([0x32, 0x33, 0x32]).then(function() { Promise.resolve(); });
				}).then(function() {
					log("prms.cashsum: "+prms.cashsum);
					const cashsum = PU.convIntToInt(prms.cashsum/100);
					log("cashsum: "+cashsum);
					log("prms.nocashsum: "+prms.nocashsum);
					const nocashsum = PU.convIntToInt(prms.nocashsum/100);
					log("nocashsum: "+nocashsum);
					const enccashier = getcashier(!!prms && !!prms.cashier ? prms.cashier: false);
					log("enccashier: "+enccashier);
					const corrcheckcmd = [0x23].concat(cp866encode("58")).concat(enccashier).concat(cashsum).concat(0x1C).concat(nocashsum).concat(0x1C).
						concat([0x30, 0x1c, 0x30, 0x1c, 0x30, 0x1c, 0x30, PU.convIntToInt(prms.op), 0x1c]);
					return talk(corrcheckcmd);
				}).then(function(data) {
					close(onfinish);
				}).catch(function(e) {
					close(err(e.message || e, e.code));
				});
			});

		}

		proc(beep, demo, info, type, check, print, z, x, changeext, cashin, cashout, ofdreportext, corrprint, hackfr);
	});
}
kkm.pirit54fz.errors = kkm.pirit.errors;


kkm.find.find = function (logging, config, proc, complete) {
	kkm.util.incl(logging());

	var types = 0;
	var ids = [];
 	var searches = 0;
 	const wd = setTimeout(function () { complete(ids); }, config().timeout);

 	function done() {
 		if (0 == --searches) {
 			clearTimeout(wd);
 			complete(ids);
 		}
 	}


	function search(enm, req, protocols, reqfail) {
		types++;
		enm(logging, config, function (n) { searches+=n; types--; if (0 == types && 0 == searches) complete(ids); }, function (confchange) {
			const conf = confchange(config);
			ids.push(conf.id);
			protocols(conf, function (conf, command, api, found, notfound) {
				var timer;
				function exec(logging, chain) {
					req(logging, conf, function (conn, send1, receive1, close, flush) {
						timer = setTimeout(function () {
							log("device not found on port "+conf.id);
							timer = false;
							close(notfound);
						}, conf.cmdtimeout);
						command(conf, chain, logging, conn, send1, receive1, close, flush);
					}, function () { done(); reqfail(conf.id); });
				}

				api(logging, conf, exec, function (beep, demo, info, type, check, print, z) {
					var typedata;
					check({}, function (res, datas) {
							if (timer)
								clearTimeout(timer);
							found(conf, res, datas);
							done();
					});
				});
			}, done);
		});
	}

	proc(search);
};


kkm.find.vendoric = function (config, logging, findlog, connector, vendor, found, notfound) {
    const L = logging();
    const log = L.log;

    return new Promise(function (res) {
    	var conns;
	    const reportcount = function(count) {
	    	conns = count;
	        log(0 == count ? "Не найдено подключений": "Найдено подключений: "+count);
	    };

    	const searches = [];
	    connector.enum(logging, config, reportcount, function (confchange) {
	    	const conconfig = confchange(config);

	    	var timeouts = {};

	        const executormaker = function(presolve) {
	        	return function(logging, check) {
		            connector.request(logging, conconfig, function(conn, send1, receive1, close, flush) {
	                    timeouts[conconfig.id] = setTimeout(function () {
	                    	close(function() {
		                        notfound(conconfig.id);
		                        findlog("На подключении "+conconfig.id+" ФР не найден");
		                        presolve({"found":false, "id":conconfig.id});
	                    	});
	                    }, conconfig.cmdtimeout);
		                vendor.command(conconfig, check, logging, conn, send1, receive1, close, flush);
		            }, function() { log(conconfig.id+" not available"); });
		        };
		    };

		    const checkconn = function(conconfig) {
		        try {
		            return new Promise(function(res) {
		                vendor.ext(logging, conconfig, executormaker(res), function(beep, demo, info, type, check) {
		                    findlog("Ищем ФР на подключении "+conconfig.id);
		                    check({}, function (fr) {
		                        clearTimeout(timeouts[conconfig.id]);
		                        found(conconfig, fr);
		                        findlog("На подключении "+conconfig.id+" обнаружен ФР "+fr.model)
		                        res({"found":true, "fr":fr});
		                    });
		                });
		            });
		        } catch (e) {
		            log("error checking port "+conconfig.id+" "+(e.message ? e.message : e));
		            findlog("Ошибка при проверке подключения "+conconfig.id);
		            return Promise.resolve({"found":false, "id":conconfig.id})
		        }
		    }

	        searches.push(checkconn(conconfig));
	        if (searches.length == conns) res(searches);
	    });
    }).then(function(searches) {
    	return Promise.race(searches);
    });
}
kkm.find.all = function (logging, findlog, found, notfound, complete) {
	var somefound = false;
	function atolfound(config, res, datas) {
		somefound = true;
		kkm.util.incl(logging());
		log("Устройство АТОЛ обнаружено на COM-порту "+config.id);
		found([
			config.id,
			res.model + " (АТОЛ)"+res.ver,
			"atol",
			false
		]);
	}

	function atol54fzfound(config, res, datas) {
		somefound = true;
		kkm.util.incl(logging());
		log("Устройство АТОЛ (54-ФЗ) обнаружено на COM-порту "+config.id);
		found([
			config.id,
			res.model + " (АТОЛ 54-ФЗ)"+res.ver,/*-*///+">"+res.ffd+">"+res.ffd105vat
			"atol54fz",
			true
		]);
	}

	function shtrikhfound(config, res, datas) {
		somefound = true;
		kkm.util.incl(logging());
		log("Устройство Штрих-М обнаружено на COM-порту "+config.id);
		found([
			config.id,
			res.model + " (Штрих-М"+(res.fz54? " 54-ФЗ": "")+")"+res.ver,
			res.fz54 ? "shtrikh54fz": "shtrikh",
			res.fz54
		]);
	}

	function piritfound(config, res, datas) {
		kkm.util.incl(logging());
		log("Устройство Pirit обнаружено на COM-порту "+config.id);
		found([
			config.id,
			res.model + " (Пирит"+(res.fz54 ? " 54-ФЗ)" : ")")+res.ver,
			res.fz54 ? "pirit54fz" : "pirit",
			res.fz54
		]);
	}

	function serialenum (logging, config, count, proc) {
		function c(num) {
			findlog(0 == num ? "Не найдено ни одного COM-порта ": "Найдено COM-портов: "+num);
			count(num);
		}

		kkm.serial.enum(logging, config, c, proc);
	}

	function bluetoothenum(logging, config, count, proc) {
		function c (num) {
			findlog("Обнаружено Bluetooth устройств: "+num);
			count(num);
		}

		kkm.bluetooth.enum(logging, config, c, proc);
	}

	function reqfail(port) {
		findlog("Порт "+port+" недоступен");
	}

	kkm.util.incl(logging());
	kkm.find.find(logging, kkm.config.default, function (search) {
		search(serialenum, kkm.serial.request, function (conf, apply, nothing) {
			findlog("Ищем устройства АТОЛ (54-ФЗ) на COM-порту "+conf.id);
			apply(conf.atol(), kkm.atol54fz.command, kkm.atol54fz.ext, atol54fzfound,
				function () {
					findlog("Ищем устройства Штрих-М на COM-порту "+conf.id);
					apply(conf.shtrikh(), kkm.shtrikh.command, kkm.shtrikh.ext, shtrikhfound,
						function () {
							findlog("Ищем устройства АТОЛ на COM-порту "+conf.id);
							apply(conf.atol(), kkm.atol.command, kkm.atol.ext, atolfound,
								function () {
					                findlog("Не удалось получить ответ от COM-порта "+conf.id);
					                nothing(); notfound(conf.id);
								}
							);
						}
					);
				}
			);
		}, reqfail);
	}, function () {
		if (somefound) return complete([]);
		setTimeout(function() {
			kkm.find.find(logging, kkm.config.default, function (search) {
				search(serialenum, kkm.serial.request, function (conf, apply, nothing) {
					findlog("Ищем устройства Pirit на COM-порту "+conf.id);
					apply(conf.pirit(), kkm.pirit.command, kkm.pirit.ext, piritfound,
		        function () {
		            findlog("Не удалось получить ответ от COM-порта "+conf.id);
		            nothing(); notfound(conf.id);
		        }
			    );
				}, reqfail);
			}, complete);
		}, 10000);
	});
}

kkm.find.forvendor = function(vendor, logging, findlog, found, notfound, complete) {
	const cd = kkm.config.default();

	function foundmaker(vendor) {
		return function(config, res) {
			switch (vendor) {
				case "atol": found([
					config.id,
					res.model + " (АТОЛ)"+res.ver,
					"atol",
					false
				]); break;
				case "atol54fz": found([
					config.id,
					res.model + " (АТОЛ 54-ФЗ)"+res.ver,
					"atol54fz",
					true
				]); break;
				case "shtrikh": found([
					config.id,
					res.model + " (Штрих-М"+(res.fz54? " 54-ФЗ": "")+")"+res.ver,
					res.fz54 ? "shtrikh54fz": "shtrikh",
					res.fz54
				]); break;
				case "pirit": found([
					config.id,
					res.model + " (Пирит"+(res.fz54 ? " 54-ФЗ)" : ")")+res.ver,
					res.fz54 ? "pirit54fz" : "pirit",
					res.fz54
				]); break;
				default: notfound(config.id);
			}
		}
	}

    function searchpromise () {
        if ("pirit" == vendor) return kkm.find.vendoric(cd.pirit, logging, findlog, kkm.serial, kkm.pirit, foundmaker("pirit"), notfound);
        if ("shtrikh" == vendor) return kkm.find.vendoric(cd.shtrikh, logging, findlog, kkm.serial, kkm.shtrikh, foundmaker("shtrikh"), notfound);
        if ("atol" == vendor) {
            const atol54fzres = kkm.find.vendoric(cd.atol54fz, logging, findlog, kkm.serial, kkm.atol54fz, foundmaker("atol54fz"), function() {/* proceeding to search non-fz */});
            return atol54fzres.then(function (ress) {
                return ress.found ?
                	Promise.resolve(ress) :
                	kkm.find.vendoric(cd.atol, logging, findlog, kkm.serial, kkm.atol, foundmaker("atol"), notfound);
            });
        }
        else return Promise.reject(new Error("Выбранный вендор не поддерживается"));
    }
	function searchtimeout() {
		return new Promise(function (r) {
			setTimeout(function() {
				findlog("Ничего не найдено за отведенное на поиск время");
				r();
			}, cd.cmdtimeout*2);
		});
	}

    Promise.race([searchpromise(), searchtimeout()]).then(complete);
}
const retail = document.getElementById("retail");
function post(data) {
    retail.contentWindow.postMessage(JSON.stringify(data), "*");
}

(resize = function () {
    retail.style.height = ""+innerHeight+"px";
})();

retail.addEventListener('permissionrequest', function(e) {
    console.log("permission request for "+e.permission);
    e.request.allow();
});
retail.addEventListener("loadstop", function () {
    post({"kkm":"init", "ver": chrome.runtime.getManifest().version});
    chrome.storage.local.get("inst", function (res) {
        if (res.inst) return;
        retail.clearData({}, {appcache:true}, function () { console.log("appcache cleared"); });
        chrome.storage.local.set({"inst":new Date().getTime()});
    })
});
retail.addEventListener("newwindow", function(e) {
    e.preventDefault();
    chrome.app.window.create("external.html?"+escape(e.targetUrl), {'state':"maximized"});
});

window.onresize = function () {
    resize();
};
const logbuf = [];
function log(msg) {
    logbuf.push(msg);
}
function flushlog() {
    const all = logbuf.join("\n");
    console.log(all);
    post({"kkm":"log", "log":all});
    logbuf.length = 0;
}

function disable (what) {
    post({"kkm":"disable", "what":what});
}

function error2retail(msg, code) {
    flushlog();
    log("sending error message ["+msg+"] to retail");
    post({"kkm":"error", "error": msg, "fatal":true, "code":(code || 999)});
    disable("print");
    disable("z");
}

function headersmap(s) {
    m = {};

    function parseline(l) {
        i = l.indexOf(':');
        m[l.substr(0, i)] = l.substr(i+1, l.length);
    }
    s.split("\n").map(parseline);

    return m;
};

window.addEventListener("message", (function (event) {
    const logging = function() {
        return {
            log: log,
            err: error2retail
        };
    };

    const U = kkm.util;
    const findlogging = U.fslogging(logging, U.id);

    var vendor = false, comm = false, config = false, executor = false, ext = false;

    function selected() {
        return (vendor && comm && config && executor && ext);
    }

    function app(proc) {
        if (selected())
            ext(logging, config, executor, proc);
        else
            error2retail("no kkm selected");
    }

    var findid = 0;

    function oncomplete (resp) {
        const res = {"kkm": "complete"};
        if (resp) {
            res.data = resp;
            log("got response data: "+JSON.stringify(resp));
        }
        post(res);
        flushlog();
    }

    const commands = {
        "beep": function (data, onfinish) { app(function (beep, demo, info, type, check, print, z) { beep(data, onfinish); }); },
        "demo": function (data, onfinish) { app(function (beep, demo, info, type, check, print, z) { demo(data, onfinish); }); },
        "print": function (data, onfinish) { log("...printing..."); app(function (beep, demo, info, type, check, print, z) { print(data, onfinish); }); },
        "corrprint": function (data, onfinish) { log("...corr check printing..."); app(function (beep, demo, info, type, check, print, z, x, change, cashin, cashout, ofdreport, corrprint) { corrprint(data, onfinish); }); },
        "z": function (data, onfinish) { log("...z..."); app(function (beep, demo, info, type, check, print, z) { z(data, function() { disable("z"); onfinish(); }); });},
        "x": function (data, onfinish) { log("...x..."); app(function (beep, demo, info, type, check, print, z, x) { x(data, onfinish); });},
        "change": function(data, onfinish) { log("...opening change..."); app(function(beep, demo, info, type, check, print, z, x, change) { change(data, onfinish); }); },
        "cashin": function(data, onfinish) { log("...cash in..."); app(function(beep, demo, info, type, check, print, z, x, change, cashin) { cashin(data, onfinish); });},
        "cashout": function(data, onfinish) { log("...cash out..."); app(function(beep, demo, info, type, check, print, z, x, change, cashin, cashout) { cashout(data, onfinish); });},
        "ofdreport": function(data, onfinish) { log("...ofd report..."); app(function(b, d, i, t, c, p, z, x, ch, cin, cout, ofdreport) { ofdreport(data, onfinish); }); },
        "check": function (data, onfinish) {
            log("checking...?");
            if (! selected()) {
                post({"kkm":"noconn"});
                log("port not specified");
                return;
            }

            log("...checking for "+vendor+"...");
            const res = {"kkm": "check"};

            app(function (beep, demo, info, type, check, print, z) {
                check(res, function (res, datas) {
                    log("got fr properties "+res);
                    log("got fr vendor "+res.vendor);
                    kkm[res.vendor].qr_supported = res.qrsupported;
                    kkm[res.vendor].ffd = res.ffd;
                    kkm[res.vendor].ffd105vat = res.ffd105vat;
						  res['version'] = chrome.runtime.getManifest().version;
                    log("check complete");
                    log("sending to app device info: ");
                    log(JSON.stringify(res));
                    post(res);
                    flushlog();
                });
            });
        },
        "printer": function(data, onfinish) {
            const fr = document.querySelector("iframe#printer");
            fr.srcdoc = data;
            fr.onload = function () { fr.contentWindow.print(); };
            onfinish();
        },
        "refresh": function (data, onfinish) {
            if (findid==100) findid = 0;
            findid++;
            const myfindid=findid;

            function findlog(msg) {
              if(findid==myfindid)
                post({"kkm":"findlog", "msg": msg});
            }

            function found(fr) {
                log("found fr "+fr[1]+" ("+fr[0]+")");
                if(findid==myfindid)
                post({"kkm":"found", "fr": fr});
            }

            function notfound(id) {
                log("not found anything on "+id);
                if(findid==myfindid)
                post({"kkm": "notfound","id": id});
            }

            function complete(ids) {
              if(findid==myfindid)
              {
                post({"kkm": "ids", "ids": ids});
                setTimeout(onfinish, 500);
              }
            }

            if (data && data.vendor) kkm.find.forvendor(data.vendor, findlogging, findlog, found, notfound, complete);
            else kkm.find.all(findlogging, findlog, found, notfound, complete);
        },
        "break": function (data, onfinish) {
          if(findid==100)
          {
            findid=0;
          }
          findid++;
            const execs = {};
            function checkcompleter (val) {
                return function () {
                    log("checking for break completed");
                    execs[val] = true;

                    var count = 0;
                    var complete = true;
                    for (k in execs) {
                        if (!execs[k]) {
                            complete = false;
                            count++;
                        }
                    }

                    log("break is "+(complete ? "complete": "incomplete, "+count+" conns more"));

                    if (complete)
                        onfinish();
                }
            }

            log("closing "+kkm.request.break.size);
            kkm.request.break.forEach(function (val, _) {
                execs[val] = false;
                val(checkcompleter(val));
            });
        },
        "select": function (data, onfinish) {
            const vc = data[2].split("_");
            vendor = vc[0];
            log("selected vendor "+vendor);
            comm = (1 < vc.length ? vc[1] : "serial");
            log("selected comm type "+comm);
            kkm.config.default.bitrate = parseInt(data[3]);
            log("speed set to "+data[3])
            ext = kkm[vendor].ext;
            config = kkm.config.default();
            config = config[vendor]();
            config.id = data[0];

            executor = function (logging, chain) {
                kkm[comm].request(logging, config, function (conn, send1, receive1, close, flush, state) {
                    kkm[vendor].command(config, chain, logging, conn, send1, receive1, close, flush, state);
                }, function () { log(config.id+" not available"); }, oncomplete);
            }
            log("fr selected on port "+data[0]);
            app(function(b, d, i, t, c, p, z, x, ch, cin, cout, ofd, corr, hackfr) { hackfr(data, onfinish); });
        },
        "setspeed": function(data, onfinish) {
            kkm.config.default.bitrate = parseInt(data);
            onfinish();
        },
        "testprint": function (data, onfinish) {
            app(function (beep, demo, info, type, check, print, z) { log("performing testprint"); demo(data, onfinish); });
        },
        "version": function (data, onfinish) {
            post({"kkm":"version", "version": chrome.runtime.getManifest().version});
            onfinish();
        },
        "relog": function (data, onfinish) {
            log(data);
        },
        "httpreq": function(data, onfinish) {
            const xhr = new XMLHttpRequest();
            xhr.onreadystatechange = function() {
                if (4 == xhr.readyState) {
                    onfinish({
                        "status": xhr.status,
                        "body": xhr.responseText,
                        "headers": headersmap(xhr.getAllResponseHeaders())
                    });
                }
            };
            xhr.open(data.method, data.url);
            for (h in data.headers) {
                xhr.setRequestHeader(h, data.headers[h]);
            }
            xhr.send(data.body);
        }
    };

    function command(name, data) {
        if (!commands[name]) return;
        commands[name](data, oncomplete);
    }

    return function (event) {
        if (event.data.lastIndexOf("MAGIC", 0) != 0) {
            log("got message " + event.data);
            try {
                const req = JSON.parse(event.data);
                log("processing with " + req.msg);
                command(req.msg, req.data);
            }
            catch (e) {
                error2retail(e.message, false);
                flushlog();
            }
        }
    }
})(), false);

chrome.app.runtime.onLaunched.addListener(function (launchData) {
    log("app launched from <"+launchData.url+">");
});
